﻿/*  Copyright 2010 Imisnew2

    http://www.TeamPlayerGaming.com/members/Imisnew2.html

    This file is part of Cameron "Imisnew2" Gunnin's Teamspeak 3 Sync Plugin
    for BFBC2 PRoCon.

    Imisnew2's Teamspeak 3 Sync Plugin for BFBC2 PRoCon is free software:
    you can redistribute it and/or modify it under the terms of the GNU
    General Public License as published by the Free Software Foundation,
    either version 3 of the License, or (at your option) any later version.

    Imisnew2's Teamspeak 3 Sync Plugin for BFBC2 PRoCon is distributed in
    the hope that it will be useful, but WITHOUT ANY WARRANTY; without even
    the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
    See the GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with Imisnew2's Teamspeak 3 Sync Plugin for BFBC2 PRoCon.
    If not, see <http://www.gnu.org/licenses/>.

*/

using System;
using System.Collections;
using System.Net.Sockets;
using System.Text;
using System.Threading;
using System.Collections.ObjectModel;
using System.Collections.Generic;

using PRoCon.Core;
using PRoCon.Core.Players;
using PRoCon.Core.Plugin;

namespace PRoConEvents
{
    /// <summary>
    /// Manages a Teamspeak 3 server, synchronizing it with a Bad Company 2 server.
    /// Moves players between specified channels corresponding to their team/squad in
    /// Bad Company 2. Also allows 'Yells' to be sent to players who are not on your
    /// Teamspeak server, yet playing in your Bad Company 2 server.
    /// </summary
    class TeamspeakSync : CPRoConMarshalByRefObject, IPRoConPluginInterface
    {
        #region stealth
        #region Whitelist Queue

        public class QQItem
        {
            private TeamspeakQuery ts3Query = null;
            private String id = null;
            private List<Object> parameters = null;

            public String Id { get { return id; } set { id = value; } }
            public TeamspeakQuery Ts3Query { get { return ts3Query; } set { ts3Query = value; } }
            public List<Object> Parameters { get { return parameters; } set { parameters = value; } }

            public QQItem(TeamspeakQuery query, String id)
            {
                Ts3Query = query;
                Id = id;
            }

            public QQItem(TeamspeakQuery query, String id, List<Object> parameters)
            {
                Ts3Query = query;
                Id = id;
                Parameters = parameters;
            }
        }

        Queue queryQueue = new Queue();
        Queue timeQueue = new Queue();

        public void checkTimeQueue()
        {
            if (timeQueue.Count == 0)
                return;
            TimeSpan t = DateTime.Now - (DateTime)timeQueue.Peek();
            if (t.TotalSeconds >= 3)
            {
                timeQueue.Dequeue();
            }
        }

        public void sendQueryAttempt()
        {
            checkTimeQueue();
            if (queryQueue.Count > 0)
            {
                if (timeQueue.Count < 9)
                {
                    QQItem queryQueueItem = (QQItem)queryQueue.Dequeue();
                    handleResponse(ts3Connection.sendQuery(queryQueueItem.Ts3Query), queryQueueItem.Id, queryQueueItem.Parameters);
                    timeQueue.Enqueue(DateTime.Now);
                }
            }
        }

        #endregion
        #region Handle Responses

        List<TeamspeakClient> ts3TempClientInfo;
        List<TeamspeakChannel> tempchannels;

        public void handleResponse(TeamspeakResponse ts3Response2, String id, List<Object> parameters)
        {
            switch (id)
            {
                case "clientlist": handleClientList(ts3Response2);
                    break;
                case "channellist": handleChannelList(ts3Response2);
                    cleanClientList();
                    requestAdvancedInfo();
                    break;
                case "advancedinfo": handleAdvancedInfo(ts3Response2, (int)parameters[0]);
                    break;
                case "swapclient": handleSwapClient(ts3Response2, (MasterClient)parameters[0]);
                    break;
                case "teamswapclient": handleSwapClientTeam(ts3Response2, (MasterClient)parameters[0]);
                    break;
                case "squadswapclient": handleSwapClientSquad(ts3Response2, (MasterClient)parameters[0]);
                    break;
                case "removeclient": handleRemoveClient(ts3Response2, (MasterClient)parameters[0]);
                    break;
            }
        }

        public void handleRemoveClient(TeamspeakResponse tsr, MasterClient client)
        {
            if (tsr.Id == 0)
                client.TsClient.medChannelId = ts3StagingChannel.tsId;
        }

        public void handleSwapClientSquad(TeamspeakResponse tsr, MasterClient client)
        {
            if (tsr.Id == 0)
                client.TsClient.medChannelId = ts3SquadChannels[client.BcClient.TeamId][client.BcClient.SquadId].tsId.Value;
            // This handles an issue where the channel in the code behind was removed
            //   when the channel in ts3 was. (768 = Invalid Channel ID)
            else if (tsr.Id == 768)
            {
                ts3SquadChannels[client.BcClient.TeamId].Remove(client.BcClient.SquadId);
                ts3Response = new TeamspeakResponse("error id=0 msg=ok");
            }
        }

        public void handleSwapClientTeam(TeamspeakResponse tsr, MasterClient client)
        {
            if (tsr.Id == 0)
                client.TsClient.medChannelId = ts3TeamChannels[client.BcClient.TeamId].tsId.Value;
            // This handles an issue where the channel in the code behind was removed
            //   when the channel in ts3 was. (768 = Invalid Channel ID)
            else if (tsr.Id == 768)
            {
                ts3TeamChannels.Remove(client.BcClient.TeamId);
                ts3SquadChannels.Remove(client.BcClient.TeamId);
                ts3Response = new TeamspeakResponse("error id=0 msg=ok");
            }
        }

        public void handleSwapClient(TeamspeakResponse tsr, MasterClient client)
        {
            if (tsr.Id == 0)
                client.TsClient.medChannelId = ts3StagingChannel.tsId;
        }

        public void handleClientList(TeamspeakResponse tsr)
        {
            ts3TempClientInfo = new List<TeamspeakClient>();
            if (tsr.Id != 0)
                return;

            // Build new teamspeak client list.
            foreach (TeamspeakResponseSection sec in tsr.Sections)
                foreach (TeamspeakResponseGroup grp in sec.Groups)
                    ts3TempClientInfo.Add(new TeamspeakClient(grp));
        }

        public void handleChannelList(TeamspeakResponse tsr)
        {
            tempchannels = new List<TeamspeakChannel>();
            if (tsr.Id != 0)
                return;

            // Build a list of channels.
            foreach (TeamspeakResponseSection section in tsr.Sections)
                foreach (TeamspeakResponseGroup group in section.Groups)
                    tempchannels.Add(new TeamspeakChannel(group));
        }

        public void requestAdvancedInfo()
        {
            if (ts3TempClientInfo == null)
                return;
            // Retrieve the response for each client.
            for (int i = 0; i < ts3TempClientInfo.Count; i++)
            {
                // Obtain the advanced information for a client.
                List<Object> tmp = new List<object>(); tmp.Add(i);
                queryQueue.Enqueue(new QQItem(TeamspeakQuery.buildClientInfoQuery(ts3TempClientInfo[i].tsId.Value), "advancedinfo", tmp));
            }
        }

        public void handleAdvancedInfo(TeamspeakResponse tsr, int arrayId)
        {
            if (arrayId < 0)
                return;
            // Check if response had connection error.
            if (tsr.Id < 0)
                return;

            // Update client.
            if (ts3TempClientInfo[arrayId] != null)
                ts3TempClientInfo[arrayId].setAdvancedData(tsr.Sections[0].Groups[0]);
            if (arrayId == ts3TempClientInfo.Count - 1)
                finalizeClientList();
        }

        public void finalizeClientList()
        {
            // Clean up clients who had faulty responses.
            for (int i = 0; i < ts3ClientInfo.Count; i++)
                if (ts3TempClientInfo[i].advIpAddress == null)
                    ts3TempClientInfo.RemoveAt(i--);

            // Replace global list with new list.
            this.ts3ClientInfo = ts3TempClientInfo;

            // Replace current response with "ok" response, so the program doesn't see
            // an issue with a possible query error when updating the clients.
            ts3Response = new TeamspeakResponse("error id=0 msg=ok");

            ts3TempClientInfo = null;
            tempchannels = null;
        }

        public void cleanClientList()
        {
            if (ts3TempClientInfo == null || tempchannels == null)
                return;
            for (int i = 0; i < ts3TempClientInfo.Count; i++)
            {
                Boolean inChannel = false;

                // Check to see if user is in staging channel.
                if (ts3TempClientInfo[i].medChannelId == ts3StagingChannel.tsId)
                    inChannel = true;
                // Check the current channel's parent to see if it is a child of the staging channel.
                // If not, check the channel's grandparent to see if it is a child of the staging channel.
                // We only check two layers deep because the plugin will never have channels three layers deep.
                foreach (TeamspeakChannel tsChannel in tempchannels)
                {
                    // Check if the user's channel is a child of the staging channel.
                    if (tsChannel.tsId == ts3TempClientInfo[i].medChannelId)
                        if (tsChannel.medPId == ts3StagingChannel.tsId)
                        {
                            inChannel = true;
                            break;
                        }
                        // Check if the user's channel is a grandchild of the staging channel.
                        else foreach (TeamspeakChannel tsChannel2 in tempchannels)
                                if (tsChannel2.tsId == tsChannel.medPId)
                                    if (tsChannel2.medPId == ts3StagingChannel.tsId)
                                    {
                                        inChannel = true;
                                        break;
                                    }

                    // Break out if the confirmed the channel as a grandchild.
                    if (inChannel)
                        break;
                }

                // Remove user if they're not in any of the channels.
                if (!inChannel)
                    ts3TempClientInfo.RemoveAt(i--);
            }
        }
        #endregion
        #endregion

        #region Teamspeak API

        #region Main Classes

        /// <summary>Holds all the information needed to connect to a TS3 server.</summary>
        public class TeamspeakConnection
        {
            private static TeamspeakResponse TSR_OK = new TeamspeakResponse("error id=0 msg=ok");
            private static TeamspeakResponse TSR_CONN_ERR_1 = new TeamspeakResponse("error id=-1 msg=Invalid\\sIP\\/Domain\\sName\\sor\\sPort.");
            private static TeamspeakResponse TSR_CONN_ERR_2 = new TeamspeakResponse("error id=-2 msg=Connection\\sto\\sthe\\sTeamspeak3\\sServer\\sfailed.");
            private static TeamspeakResponse TSR_CONN_ERR_3 = new TeamspeakResponse("error id=-3 msg=Server\\swas\\snot\\sa\\sTeamspeak3\\sServer.");
            private static TeamspeakResponse TSR_QUER_ERR_1 = new TeamspeakResponse("error id=-4 msg=No\\sConnection\\sis\\sestablished.");
            private static TeamspeakResponse TSR_QUER_ERR_2 = new TeamspeakResponse("error id=-5 msg=An\\serror\\soccurred\\swhen\\ssending\\sa\\squery.");
            private static TeamspeakResponse TSR_QUER_ERR_3 = new TeamspeakResponse("error id=-6 msg=An\\serror\\soccurred\\swhen\\sreceiving\\sa\\sresponse.");

            // [0] The Socket to which the connection is bound to.
            // [1] The Ip/Host Name to use when connecting to the Teamspeak 3 server.
            // [2] The Port to use when connecting to the Teamspeak 3 server.
            private TcpClient tsSocket = null;
            private String tsIp = null;
            private UInt16 tsPort = 0;

            /// <summary>The Ip/Host Name to use when connecting to the Teamspeak 3 server.</summary>
            public String IP { get { return tsIp; } set { tsIp = value; } }
            /// <summary>The Port to use when connecting to the Teamspeak 3 server.</summary>
            public UInt16 Port { get { return tsPort; } set { tsPort = value; } }
            /// <summary>Whether the connection is established.</summary>
            public Boolean IsConnected { get { return (tsSocket != null && tsSocket.Connected); } }

            /// <summary>Creates the TeamspeakConnection object.</summary>
            /// <param name="ip">The IP or Domain Name of the Teamspeak3 Server.</param>
            /// <param name="port">The Query Port of the Teamspeak3 Server.</param>
            public TeamspeakConnection(String ip, UInt16 port)
            {
                IP = ip;
                Port = port;
            }

            #region Connection Methods

            /// <summary>Establish a connection to a Teamspeak 3 Server. This is a blocking call.</summary>
            /// <returns>TeamspeakResponse - How the method performed.</returns>
            public TeamspeakResponse openConnection()
            {
                // If a connection is already established, close it first.
                if (tsSocket != null)
                    closeConnection();

                // Check the IP and Port to make sure they are valid.
                if (tsIp == null || tsPort == 0)
                    return TSR_CONN_ERR_1;

                // Attempt to establish a connection.
                try { tsSocket = new TcpClient(tsIp, tsPort); }
                catch (Exception) { tsSocket = null; return TSR_CONN_ERR_2; }

                // Setup a timeout for read/write errors.
                tsSocket.GetStream().ReadTimeout = 5000;
                tsSocket.GetStream().WriteTimeout = 5000;

                // Read in the first response to make sure the connection was to a Teamspeak 3 server.
                Byte[] streamBuffer = new Byte[512];
                try
                {
                    Int32 size = tsSocket.GetStream().Read(streamBuffer, 0, streamBuffer.Length);
                    if (Encoding.Default.GetString(streamBuffer, 0, size).Split('\n')[0] != "TS3")
                    { closeConnection(); return TSR_CONN_ERR_3; }
                }
                catch (Exception) { closeConnection(); return TSR_CONN_ERR_2; }

                // If everything has worked up to this point, return a successful connection.
                return TSR_OK;
            }

            /// <summary>Closes the connection to a Teamspeak 3 Server.</summary>
            /// <returns>TeamspeakResponse - How the method performed.</returns>
            public TeamspeakResponse closeConnection()
            {
                if (tsSocket != null)
                {
                    if (tsSocket.Connected)
                        tsSocket.Close();
                    tsSocket = null;
                }
                return TSR_OK;
            }

            #endregion

            #region Send/Receive Methods

            /// <summary>Sends a query and blocks until a response is received.</summary>
            /// <param name="query">The query to send to the server.</param>
            /// <returns>The response from the server.</returns>
            public TeamspeakResponse sendQuery(TeamspeakQuery query)
            {
                String responseBuffer = String.Empty;
                Byte[] streamBuffer = new Byte[65536];

                // Only send a query if we're connected.
                if (tsSocket == null)
                    return TSR_QUER_ERR_1;

                // Send the query.
                try
                {
                    streamBuffer = Encoding.Default.GetBytes(query.rawQuery());
                    tsSocket.GetStream().Write(streamBuffer, 0, streamBuffer.Length);
                }
                catch (Exception) { closeConnection(); return TSR_QUER_ERR_2; }

                // Receive the response.
                while (!responseBuffer.Contains("error id=") || !responseBuffer.EndsWith("\n\r"))
                    try
                    {
                        Int32 size = tsSocket.GetStream().Read(streamBuffer, 0, streamBuffer.Length);
                        responseBuffer += Encoding.Default.GetString(streamBuffer, 0, size);
                    }
                    catch (Exception) { closeConnection(); return TSR_QUER_ERR_3; }

                // Send back the response.
                return new TeamspeakResponse(responseBuffer);
            }

            #endregion
        }

        /// <summary>Contains all the information possible in a response from a Teamspeak 3 server.</summary>
        public class TeamspeakResponse
        {
            // [0] The raw response.
            // [1] The id of the response.
            // [2] The message of the response.
            private String tsResponse = null;
            private Int32? tsId = null;
            private String tsMessage = null;

            // Each section of the response.  Divided by new lines.
            //  Contains: Each Group of the section.  Divided by pipes.
            //    Contains: Each Key/Value pair of the group.  Divided by equals signs.
            private List<TeamspeakResponseSection> tsSections = null;

            /// <summary>The raw response from the server.</summary>
            public String RawResponse { get { return tsResponse; } }
            /// <summary>The Id for this response.</summary>
            public Int32? Id { get { return tsId; } }
            /// <summary>The Message for this response.</summary>
            public String Message { get { return tsMessage; } }
            /// <summary>Whether there are sections in this response.</summary>
            public Boolean HasSections { get { return tsSections.Count != 0; } }
            /// <summary>A Read-only list of sections found in the response.</summary>
            public ReadOnlyCollection<TeamspeakResponseSection> Sections { get { return tsSections.AsReadOnly(); } }

            /// <summary>Create the object and parse the raw response into sections.</summary>
            /// <param name="rawResponse">The raw response from the server.</param>
            public TeamspeakResponse(String rawResponse) { parse(rawResponse); }

            /// <summary>Parses the response into a list of sections.</summary>
            /// <param name="raw">The raw server response</param>
            private void parse(string raw)
            {
                // Set Class Variables.
                tsResponse = raw.Replace("\n\r", "\n");
                tsSections = new List<TeamspeakResponseSection>();

                // Split the response up into sections and remove invalid lines.
                foreach (String section in tsResponse.Split('\n'))
                    if (section.Contains("error id="))
                    {
                        int idStart = section.IndexOf("id=") + 3;
                        int msgStart = section.IndexOf("msg=") + 4;

                        int outInt = 0;
                        if (int.TryParse(section.Substring(idStart, msgStart - idStart - 5), out outInt))
                            tsId = outInt;
                        tsMessage = TeamspeakHelper.ts_UnescapeString(section.Substring(msgStart));
                    }
                    else if (section.Contains("="))
                        tsSections.Add(new TeamspeakResponseSection(section.Trim()));
            }
        }

        /// <summary>Contains each section's groups, which are divided by pipes.</summary>
        public class TeamspeakResponseSection
        {
            // The raw section response.
            private String tsSection = null;

            // Each group of the section.  Divided by new pipes.
            //  Contains: Each Key/Value pair of the group.  Divided by equals signs.
            private List<TeamspeakResponseGroup> tsGroups = new List<TeamspeakResponseGroup>();

            /// <summary>The raw section response.</summary>
            public String RawSection { get { return tsSection; } }
            /// <summary>Whether this section has any groups.</summary>
            public Boolean HasGroups { get { return tsGroups.Count != 0; } }
            /// <summary>A Read-only list of groups found in the response.</summary>
            public ReadOnlyCollection<TeamspeakResponseGroup> Groups { get { return tsGroups.AsReadOnly(); } }

            /// <summary>Parses the section into a list of groups.</summary>
            /// <param name="raw">The raw section response</param>
            public TeamspeakResponseSection(String rawSection) { parse(rawSection); }

            /// <summary>Parses the section into a list of groups.</summary>
            /// <param name="raw">The raw server response</param>
            private void parse(String raw)
            {
                // Set Class Variables.
                tsSection = raw;
                tsGroups = new List<TeamspeakResponseGroup>();

                // Split the section up into groups.
                foreach (String group in raw.Split('|'))
                    tsGroups.Add(new TeamspeakResponseGroup(group.Trim()));
            }
        }

        /// <summary>Contains each group's key/value pairs, which are divided by spaces then separated by equals signs.</summary>
        public class TeamspeakResponseGroup
        {
            // The raw group response.
            private String tsGroup = null;

            // Each Key/Value pair of the group.  Divided by equals signs.
            Dictionary<String, String> tsPairs = new Dictionary<String, String>();

            /// <summary>The raw section response.</summary>
            public String RawGroup { get { return tsGroup; } }
            /// <summary>Gets a single value from the dictionary of key value pairs.</summary>
            /// <param name="key">The key to search for in the dictionary.</param>
            /// <returns>The value if the key is in the dictionary.  Null otherwise.</returns>
            public String this[String key] { get { return (tsPairs.ContainsKey(key)) ? tsPairs[key] : null; } }

            /// <summary>Parse the group into a dictionary of key/value pairs.</summary>
            /// <param name="rawGroup">The raw group response.</param>
            public TeamspeakResponseGroup(String rawGroup) { parse(rawGroup); }

            /// <summary>Parses the group into a dictionary of key/value pairs.</summary>
            /// <param name="raw"></param>
            private void parse(String raw)
            {
                // Set Class Variables.
                tsGroup = raw;
                tsPairs = new Dictionary<String, String>();

                // Split the group up into key/value pairs and discard invalid pairs.
                foreach (string pair in raw.Split(' '))
                    if (pair.Contains("="))
                        // For some reason, a key is being received twice for the same response.
                        // This causes the program to crash.  To work around this, if a key is received
                        // twice, it simply stores the last value received.
                        // NOTE: The reason for receiving a key twice is unknown.  It may be due to my
                        // result parsing logic, or due to bad connection issues.  Either way, it needs
                        // to handle it somehow.  I've never actually come across this error... (perhaps
                        // because I'm on a back-bone connection?)
                        if (tsPairs.ContainsKey(pair.Split('=')[0]))
                            tsPairs[pair.Split('=')[0]] = TeamspeakHelper.ts_UnescapeString(pair.Split('=')[1]);
                        else
                            tsPairs.Add(pair.Split('=')[0], TeamspeakHelper.ts_UnescapeString(pair.Split('=')[1]));
            }
        }

        /// <summary>Holds all the information needed to send a message to teamspeak.</summary>
        public class TeamspeakQuery
        {
            // [0] The base command.
            // [1] The Key/Value pairs for the command.
            // [2] The Options for the command.
            private String tsCommand = null;
            private Dictionary<String, String> tsParameters = null;
            private List<String> tsOptions = null;

            /// <summary>Gets the command for this query.</summary>
            public String Command { get { return tsCommand; } }

            /// <summary>Creates the object and sets the command.</summary>
            /// <param name="command">The command to build.</param>
            public TeamspeakQuery(String command)
            {
                tsCommand = command;
                tsParameters = new Dictionary<String, String>();
                tsOptions = new List<String>();
            }

            /// <summary>Adds a key/value pair to this query.</summary>
            /// <param name="key">The key of the pair.</param>
            /// <param name="value">The value of the pair.</param>
            public void addParameter(String key, String value)
            {
                String tKey = key.Trim();
                String tValue = value.Trim();
                if (!String.IsNullOrEmpty(tKey) && !String.IsNullOrEmpty(tValue) && !tsParameters.ContainsKey(tKey))
                    tsParameters.Add(TeamspeakHelper.ts_EscapeString(tKey), TeamspeakHelper.ts_EscapeString(tValue));
            }

            /// <summary>Removes a key/value pair from this query.</summary>
            /// <param name="key">The key of the pair to remove.</param>
            public void removeParameter(String key)
            {
                String tKey = key.Trim();
                if (tsParameters.ContainsKey(tKey))
                    tsParameters.Remove(tKey);
            }

            /// <summary>Adds an option to this query.</summary>
            /// <param name="option">The option to add.</param>
            public void addOption(String option)
            {
                String tOption = option.Trim();
                if (!String.IsNullOrEmpty(tOption))
                    tsOptions.Add(TeamspeakHelper.ts_EscapeString(tOption));
            }

            /// <summary>Removes an option from this query.</summary>
            /// <param name="option">The option to remove.</param>
            public void removeOption(String option)
            {
                String tOption = option.Trim();
                if (tsOptions.Contains(tOption))
                    tsOptions.Remove(tOption);
            }

            /// <summary>Gets the raw query string for this query.</summary>
            /// <returns>The raw query string.</returns>
            public String rawQuery()
            {
                StringBuilder rawQuery = new StringBuilder();

                // Append the command.
                rawQuery.Append(tsCommand);

                // Append the parameters.
                foreach (KeyValuePair<String, String> p in tsParameters)
                    rawQuery.AppendFormat(" {0}={1}", p.Key, p.Value);

                // Append the options.
                foreach (String o in tsOptions)
                    rawQuery.AppendFormat(" -{0}", o);

                // Append a new line.
                rawQuery.Append("\n");

                return rawQuery.ToString();
            }


            #region Pre-Built Queries

            /// <summary>Builds a login query for you.</summary>
            /// <param name="username">The username to use.</param>
            /// <param name="password">The password to use.</param>
            public static TeamspeakQuery buildLoginQuery(String username, String password)
            {
                TeamspeakQuery tsLogin = new TeamspeakQuery("login");
                tsLogin.addParameter("client_login_name", username);
                tsLogin.addParameter("client_login_password", password);
                return tsLogin;
            }

            /// <summary>Builds a serverlist query for you.</summary>
            public static TeamspeakQuery buildServerListQuery()
            {
                return new TeamspeakQuery("serverlist");
            }

            /// <summary>Builds a use query for you.</summary>
            /// <param name="virtualId">The virtual server Id to use.</param>
            public static TeamspeakQuery buildUseVIdQuery(Int32 virtualId)
            {
                TeamspeakQuery tsUse = new TeamspeakQuery("use");
                tsUse.addParameter("sid", virtualId.ToString());
                return tsUse;
            }

            /// <summary>Builds a use query for you.</summary>
            /// <param name="port">The port of the virtual server to use.</param>
            public static TeamspeakQuery buildUsePortQuery(Int32 port)
            {
                TeamspeakQuery tsUse = new TeamspeakQuery("use");
                tsUse.addParameter("port", port.ToString());
                return tsUse;
            }

            /// <summary>Builds a channellist query for you.</summary>
            public static TeamspeakQuery buildChannelListQuery()
            {
                return new TeamspeakQuery("channellist");
            }

            /// <summary>Builds a channelfind query for you.</summary>
            /// <param name="channelName">The channel name to use.</param>
            public static TeamspeakQuery buildChannelFindQuery(String channelName)
            {
                TeamspeakQuery tsChannelFind = new TeamspeakQuery("channelfind");
                tsChannelFind.addParameter("pattern", channelName);
                return tsChannelFind;
            }

            /// <summary>Builds a channelinfo query for you.</summary>
            /// <param name="channelId">The channel id to use.</param>
            public static TeamspeakQuery buildChannelInfoQuery(Int32 channelId)
            {
                TeamspeakQuery tsChannelInfo = new TeamspeakQuery("channelinfo");
                tsChannelInfo.addParameter("cid", channelId.ToString());
                return tsChannelInfo;
            }

            /// <summary>Builds a clientlist query for you.</summary>
            public static TeamspeakQuery buildClientListQuery()
            {
                return new TeamspeakQuery("clientlist");
            }

            /// <summary>Builds a clientfind query for you.</summary>
            /// <param name="clientName">The client name to use.</param>
            public static TeamspeakQuery buildClientFindQuery(String clientName)
            {
                TeamspeakQuery tsClientFind = new TeamspeakQuery("clientfind");
                tsClientFind.addParameter("pattern", clientName);
                return tsClientFind;
            }

            /// <summary>Builds a clientinfo query for you.</summary>
            /// <param name="clientId">The client id to use.</param>
            public static TeamspeakQuery buildClientInfoQuery(Int32 clientId)
            {
                TeamspeakQuery tsClientInfo = new TeamspeakQuery("clientinfo");
                tsClientInfo.addParameter("clid", clientId.ToString());
                return tsClientInfo;
            }

            /// <summary>Builds a clientmove query for you.</summary>
            /// <param name="clientId">The client id to use.</param>
            /// <param name="channelId">The channel id to use.</param>
            public static TeamspeakQuery buildClientMoveQuery(Int32 clientId, Int32 channelId)
            {
                TeamspeakQuery tsClientMove = new TeamspeakQuery("clientmove");
                tsClientMove.addParameter("clid", clientId.ToString());
                tsClientMove.addParameter("cid", channelId.ToString());
                return tsClientMove;
            }

            #endregion
        }

        #endregion
        #region Objects

        /// <summary>Holds all possible information about a server.</summary>
        public class TeamspeakServer
        {
            #region Basic Server Data

            // Data received from a "serverlist" command
            //-- Identifying Data
            public String tsName = null; //virtualserver_name
            public Int32? tsId = null; //virtualserver_id
            public Int32? tsPort = null; //virtualserver_port
            public Int32? tsMachineId = null; //virtualserver_machine_id
            //-- Status Data
            public String tsStatus = null; //virtualserver_status
            public Int32? tsUpTime = null; //virtualserver_uptime
            public Int32? tsClientsOnline = null; //virtualserver_clientsonline
            public Int32? tsQueryClientsOnline = null; //virtualserver_queryclientsonline
            //-- Misc Data
            public Int32? tsQueryMaxClients = null; //virtualserver_maxclients
            public Boolean? tsAutoStart = null; //virtualserver_autostart

            #endregion

            /// <summary>Default constructor.</summary>
            public TeamspeakServer() { }
            /// <summary>Attempts to set all the data from the passed in response.</summary>
            /// <param name="serverInfo">The server's information to set from.</param>
            public TeamspeakServer(TeamspeakResponseGroup serverInfo)
            {
                setBasicData(serverInfo);
            }

            /// <summary>Sets all the basic data for a Teamspeak server.</summary>
            /// <param name="serverInfo">The server's information to set from.</param>
            public void setBasicData(TeamspeakResponseGroup serverInfo)
            {
                String value;
                Int32 iValue;
                Boolean bValue;

                tsName = serverInfo["virtualserver_name"];
                if ((value = serverInfo["virtualserver_id"]) != null) if (Int32.TryParse(value, out iValue)) tsId = iValue; else tsId = null; else tsId = null;
                if ((value = serverInfo["virtualserver_port"]) != null) if (Int32.TryParse(value, out iValue)) tsPort = iValue; else tsPort = null; else tsPort = null;
                if ((value = serverInfo["virtualserver_machine_id"]) != null) if (Int32.TryParse(value, out iValue)) tsMachineId = iValue; else tsMachineId = null; else tsMachineId = null;

                tsStatus = serverInfo["virtualserver_status"];
                if ((value = serverInfo["virtualserver_uptime"]) != null) if (Int32.TryParse(value, out iValue)) tsUpTime = iValue; else tsUpTime = null; else tsUpTime = null;
                if ((value = serverInfo["virtualserver_clientsonline"]) != null) if (Int32.TryParse(value, out iValue)) tsClientsOnline = iValue; else tsClientsOnline = null; else tsClientsOnline = null;
                if ((value = serverInfo["virtualserver_queryclientsonline"]) != null) if (Int32.TryParse(value, out iValue)) tsQueryClientsOnline = iValue; else tsQueryClientsOnline = null; else tsQueryClientsOnline = null;

                if ((value = serverInfo["virtualserver_maxclients"]) != null) if (Int32.TryParse(value, out iValue)) tsQueryMaxClients = iValue; else tsQueryMaxClients = null; else tsQueryMaxClients = null;
                if ((value = serverInfo["virtualserver_autostart"]) != null) if (Boolean.TryParse(value, out bValue)) tsAutoStart = bValue; else tsAutoStart = null; else tsAutoStart = null;
            }
        }

        /// <summary>Holds all possible information about a channel.</summary>
        public class TeamspeakChannel
        {
            #region Basic Channel Data

            // Data received from a "channelfind" command
            public String tsName = null;   //channel_name
            public Int32? tsId = null;   //cid

            #endregion
            #region Medium Channel Data

            // Data received from a "channellist" command
            public Int32? medPId = null;   //pid
            public Int32? medOrder = null;   //channel_order
            public Int32? medTotalClients = null;   //total_clients
            public Int32? medPowerNeededToSub = null;   //channel_needed_subscribe_power

            #endregion
            #region Advanced Channel Data

            // Additional Data received from a "channelinfo" command
            public String advTopic = null;  //channel_topic
            public String advDescription = null;  //channel_description
            public String advPassword = null;  //channel_password
            public String advFilepath = null;  //channel_filepath
            public String advPhoneticName = null;  //channel_name_phonetic
            public Int32? advCodec = null;  //channel_codec
            public Int32? advCodecQuality = null;  //channel_codec_quality
            public Int32? advCodecLatencyFactor = null;  //channel_codec_latency_factor
            public Int32? advMaxClients = null;  //channel_maxclients
            public Int32? advMaxFamilyClients = null;  //channel_maxfamilyclients
            public Int32? advNeededTalkPower = null;  //channel_needed_talk_power
            public Int32? advIconId = null;  //channel_icon_id
            public Boolean? advFlagPermanent = null;  //channel_flag_permanent
            public Boolean? advFlagSemiPermanent = null;  //channel_flag_semi_permanent
            public Boolean? advFlagDefault = null;  //channel_flag_default
            public Boolean? advFlagPassword = null;  //channel_flag_password
            public Boolean? advFlagMaxClientsUnlimited = null;  //channel_flag_maxclients_unlimited
            public Boolean? advFlagMaxFamilyClientsUnlimited = null;  //channel_flag_maxfamilyclients_unlimited
            public Boolean? advFlagMaxFamilyClientsInherited = null;  //channel_flag_maxfamilyclients_inherited
            public Boolean? advForcedSilence = null;  //channel_forced_silence

            #endregion

            /// <summary>Default constructor.</summary>
            public TeamspeakChannel() { }
            /// <summary>Attempts to set all the data from the passed in response.</summary>
            /// <param name="channelInfo">The channel's information to set from.</param>
            public TeamspeakChannel(TeamspeakResponseGroup channelInfo)
            {
                setBasicData(channelInfo);
                setMediumData(channelInfo);
                setAdvancedData(channelInfo);
            }

            /// <summary>Sets all the basic data for a Teamspeak channel.</summary>
            /// <param name="channelInfo">The channel's information to set from.</param>
            public void setBasicData(TeamspeakResponseGroup channelInfo)
            {
                String value;
                Int32 iValue;

                tsName = channelInfo["channel_name"];
                if ((value = channelInfo["cid"]) != null) if (Int32.TryParse(value, out iValue)) tsId = iValue; else tsId = null; else tsId = null;
            }

            /// <summary>Sets all the medium data for a Teamspeak channel.</summary>
            /// <param name="channelInfo">The channel's information to set from.</param>
            public void setMediumData(TeamspeakResponseGroup channelInfo)
            {
                String value;
                Int32 iValue;

                if ((value = channelInfo["pid"]) != null) if (Int32.TryParse(value, out iValue)) medPId = iValue; else medPId = null; else medPId = null;
                if ((value = channelInfo["channel_order"]) != null) if (Int32.TryParse(value, out iValue)) medOrder = iValue; else medOrder = null; else medOrder = null;
                if ((value = channelInfo["total_clients"]) != null) if (Int32.TryParse(value, out iValue)) medTotalClients = iValue; else medTotalClients = null; else medTotalClients = null;
                if ((value = channelInfo["channel_needed_subscribe_power"]) != null) if (Int32.TryParse(value, out iValue)) medPowerNeededToSub = iValue; else medPowerNeededToSub = null; else medPowerNeededToSub = null;
            }

            /// <summary>Sets all the advanced data for a Teamspeak channel.</summary>
            /// <param name="channelInfo">The channel's information to set from.</param>
            public void setAdvancedData(TeamspeakResponseGroup channelInfo)
            {
                String value;
                Int32 iValue;
                Boolean bValue;

                advTopic = channelInfo["channel_topic"];
                advDescription = channelInfo["channel_description"];
                advPassword = channelInfo["channel_password"];
                advFilepath = channelInfo["channel_filepath"];
                advPhoneticName = channelInfo["channel_name_phonetic"];
                if ((value = channelInfo["channel_codec"]) != null) if (Int32.TryParse(value, out iValue)) advCodec = iValue; else advCodec = null; else advCodec = null;
                if ((value = channelInfo["channel_codec_quality"]) != null) if (Int32.TryParse(value, out iValue)) advCodecQuality = iValue; else advCodecQuality = null; else advCodecQuality = null;
                if ((value = channelInfo["channel_codec_latency_factor"]) != null) if (Int32.TryParse(value, out iValue)) advCodecLatencyFactor = iValue; else advCodecLatencyFactor = null; else advCodecLatencyFactor = null;
                if ((value = channelInfo["channel_maxclients"]) != null) if (Int32.TryParse(value, out iValue)) advMaxClients = iValue; else advMaxClients = null; else advMaxClients = null;
                if ((value = channelInfo["channel_maxfamilyclients"]) != null) if (Int32.TryParse(value, out iValue)) advMaxFamilyClients = iValue; else advMaxFamilyClients = null; else advMaxFamilyClients = null;
                if ((value = channelInfo["channel_needed_talk_power"]) != null) if (Int32.TryParse(value, out iValue)) advNeededTalkPower = iValue; else advNeededTalkPower = null; else advNeededTalkPower = null;
                if ((value = channelInfo["channel_icon_id"]) != null) if (Int32.TryParse(value, out iValue)) advIconId = iValue; else advIconId = null; else advIconId = null;
                if ((value = channelInfo["channel_flag_permanent"]) != null) if (Boolean.TryParse(value, out bValue)) advFlagPermanent = bValue; else advFlagPermanent = null; else advFlagPermanent = null;
                if ((value = channelInfo["channel_flag_semi_permanent"]) != null) if (Boolean.TryParse(value, out bValue)) advFlagSemiPermanent = bValue; else advFlagSemiPermanent = null; else advFlagSemiPermanent = null;
                if ((value = channelInfo["channel_flag_default"]) != null) if (Boolean.TryParse(value, out bValue)) advFlagDefault = bValue; else advFlagDefault = null; else advFlagDefault = null;
                if ((value = channelInfo["channel_flag_password"]) != null) if (Boolean.TryParse(value, out bValue)) advFlagPassword = bValue; else advFlagPassword = null; else advFlagPassword = null;
                if ((value = channelInfo["channel_flag_maxclients_unlimited"]) != null) if (Boolean.TryParse(value, out bValue)) advFlagMaxClientsUnlimited = bValue; else advFlagMaxClientsUnlimited = null; else advFlagMaxClientsUnlimited = null;
                if ((value = channelInfo["channel_flag_maxfamilyclients_unlimited"]) != null) if (Boolean.TryParse(value, out bValue)) advFlagMaxFamilyClientsUnlimited = bValue; else advFlagMaxFamilyClientsUnlimited = null; else advFlagMaxFamilyClientsUnlimited = null;
                if ((value = channelInfo["channel_flag_maxfamilyclients_inherited"]) != null) if (Boolean.TryParse(value, out bValue)) advFlagMaxFamilyClientsInherited = bValue; else advFlagMaxFamilyClientsInherited = null; else advFlagMaxFamilyClientsInherited = null;
                if ((value = channelInfo["channel_forced_silence"]) != null) if (Boolean.TryParse(value, out bValue)) advForcedSilence = bValue; else advForcedSilence = null; else advForcedSilence = null;
            }
        }

        /// <summary>Holds advll possible information advbout adv client.</summary>
        public class TeamspeakClient
        {
            #region Basic Client Data

            // Data received from a "clientfind" command
            public String tsName = null;  //client_nickname
            public Int32? tsId = null;  //clid

            #endregion
            #region Medium Client Data

            // Data received from a "clientlist" command
            public Int32? medDatabaseId = null;  //client_database_id
            public Int32? medChannelId = null;  //cid
            public Int32? medType = null;  //client_type

            #endregion
            #region Advanced Client Data

            // Additional Data received from a "clientinfo" command
            //-- Identifying Data
            public String advLoginName = null; //client_login_name
            public String advUniqueId = null; //client_unique_identifier
            public String advIpAddress = null; //connection_client_ip
            //-- Meta Data
            public String advVersion = null; //client_version
            public String advPlatform = null; //client_platform
            public String advDescription = null; //client_description
            public String advCountry = null; //client_country
            public String advMetaData = null; //client_meta_data
            //-- Permissions Data
            public Int32? advChannelGroupId = null; //client_channel_group_id
            public Int32? advServerGroupId = null; //client_servergroups
            public Boolean? advIsChannelCommander = null; //client_is_channel_commander
            //-- Server Statistics Data
            public String advDefaultChannel = null; //client_default_channel
            public Int32? advConnectionTime = null; //connection_connected_time
            public Int32? advIdleTime = null; //client_idle_time
            public Int32? advCreationTime = null; //client_created
            public Int32? advLastConnected = null; //client_lastconnected
            public Int32? advTotalConnections = null; //client_totalconnections
            //-- Microphone Data
            public Boolean? advInputMuted = null; //client_input_muted
            public Boolean? advOutputMuted = null; //client_output_muted
            public Boolean? advOutputMutedOnly = null; //client_outputonly_muted
            public Boolean? advInputHardware = null; //client_input_hardware
            public Boolean? advOutputHardware = null; //client_output_hardware
            public Boolean? advIsRecording = null; //client_is_recording
            //-- Misc Data
            public String advFlagAvatar = null; //client_flag_avatar
            public String advAwayMessage = null; //client_away_message
            public String advTalkMessage = null; //client_talk_request_msg
            public String advPhoneticNick = null; //client_nickname_phonetic
            public String advDefaultToken = null; //client_default_token
            public String advBase64Hash = null; //client_base64HashClientUID
            public Int32? advTalkPower = null; //client_talk_power
            public Int32? advQueryViewPower = null; //client_needed_serverquery_view_power
            public Int32? advUnreadMessages = null; //client_unread_messages
            public Int32? advIconId = null; //client_icon_id
            public Boolean? advIsAway = null; //client_away
            public Boolean? advTalkRequest = null; //client_talk_request
            public Boolean? advIsTalker = null; //client_is_talker
            public Boolean? advIsPriority = null; //client_is_priority_speaker
            //-- Bandwidth Data
            public Int32? advBytesUpMonth = null; //client_month_bytes_uploaded
            public Int32? advBytesDownMonth = null; //client_month_bytes_downloaded
            public Int32? advBytesUpTotal = null; //client_total_bytes_uploaded
            public Int32? advBytesDownTotal = null; //client_total_bytes_downloaded
            public Int32? advFileBandwidthSent = null; //connection_filetransfer_bandwidth_sent
            public Int32? advFileBandwidthRec = null; //connection_filetransfer_bandwidth_received
            public Int32? advPacketsTotalSent = null; //connection_packets_sent_total
            public Int32? advPacketsTotalRec = null; //connection_packets_received_total
            public Int32? advBytesTotalSent = null; //connection_bytes_sent_total
            public Int32? advBytesTotalRec = null; //connection_bytes_received_total
            public Int32? advBndwdthSecondSent = null; //connection_bandwidth_sent_last_second_total
            public Int32? advBndwdthSecondRec = null; //connection_bandwidth_received_last_second_total
            public Int32? advBndwdthMinuteSent = null; //connection_bandwidth_sent_last_minute_total
            public Int32? advBndwdthMinuteRec = null; //connection_bandwidth_received_last_minute_total

            #endregion

            /// <summary>Default constructor.</summary>
            public TeamspeakClient() { }
            /// <summary>Attempts to set all the data from the passed in response.</summary>
            /// <param name="clientInfo">The client's information to set from.</param>
            public TeamspeakClient(TeamspeakResponseGroup clientInfo)
            {
                setBasicData(clientInfo);
                setMediumData(clientInfo);
                setAdvancedData(clientInfo);
            }

            /// <summary>Sets all the basic data for a Teamspeak client.</summary>
            /// <param name="clientInfo">The client's information to set from.</param>
            public void setBasicData(TeamspeakResponseGroup clientInfo)
            {
                String value;
                Int32 iValue;

                tsName = clientInfo["client_nickname"];
                if ((value = clientInfo["clid"]) != null) if (Int32.TryParse(value, out iValue)) tsId = iValue; else tsId = null; else tsId = null;
            }

            /// <summary>Sets all the medium data for a Teamspeak client.</summary>
            /// <param name="clientInfo">The client's information to set from.</param>
            public void setMediumData(TeamspeakResponseGroup clientInfo)
            {
                String value;
                Int32 iValue;

                if ((value = clientInfo["client_database_id"]) != null) if (Int32.TryParse(value, out iValue)) medDatabaseId = iValue; else medDatabaseId = null; else medDatabaseId = null;
                if ((value = clientInfo["cid"]) != null) if (Int32.TryParse(value, out iValue)) medChannelId = iValue; else medChannelId = null; else medChannelId = null;
                if ((value = clientInfo["client_type"]) != null) if (Int32.TryParse(value, out iValue)) medType = iValue; else medType = null; else medType = null;
            }

            /// <summary>Sets all the advanced data for a Teamspeak client.</summary>
            /// <param name="clientInfo">The client's information to set from.</param>
            public void setAdvancedData(TeamspeakResponseGroup clientInfo)
            {
                String value;
                Int32 iValue;
                Boolean bValue;

                advLoginName = clientInfo["client_login_name"];
                advUniqueId = clientInfo["client_unique_identifier"];
                advIpAddress = clientInfo["connection_client_ip"];

                advVersion = clientInfo["client_version"];
                advPlatform = clientInfo["client_platform"];
                advDescription = clientInfo["client_description"];
                advCountry = clientInfo["client_country"];
                advMetaData = clientInfo["client_meta_data"];

                if ((value = clientInfo["client_channel_group_id"]) != null) if (Int32.TryParse(value, out iValue)) advChannelGroupId = iValue; else advChannelGroupId = null; else advChannelGroupId = null;
                if ((value = clientInfo["client_servergroups"]) != null) if (Int32.TryParse(value, out iValue)) advServerGroupId = iValue; else advServerGroupId = null; else advServerGroupId = null;
                if ((value = clientInfo["client_is_channel_commander"]) != null) if (Boolean.TryParse(value, out bValue)) advIsChannelCommander = bValue; else advIsChannelCommander = null; else advIsChannelCommander = null;

                advDefaultChannel = clientInfo["client_default_channel"];
                if ((value = clientInfo["connection_connected_time"]) != null) if (Int32.TryParse(value, out iValue)) advConnectionTime = iValue; else advConnectionTime = null; else advConnectionTime = null;
                if ((value = clientInfo["client_idle_time"]) != null) if (Int32.TryParse(value, out iValue)) advIdleTime = iValue; else advIdleTime = null; else advIdleTime = null;
                if ((value = clientInfo["client_created"]) != null) if (Int32.TryParse(value, out iValue)) advCreationTime = iValue; else advCreationTime = null; else advCreationTime = null;
                if ((value = clientInfo["client_lastconnected"]) != null) if (Int32.TryParse(value, out iValue)) advLastConnected = iValue; else advLastConnected = null; else advLastConnected = null;
                if ((value = clientInfo["client_totalconnections"]) != null) if (Int32.TryParse(value, out iValue)) advTotalConnections = iValue; else advTotalConnections = null; else advTotalConnections = null;

                if ((value = clientInfo["client_input_muted"]) != null) if (Boolean.TryParse(value, out bValue)) advInputMuted = bValue; else advInputMuted = null; else advInputMuted = null;
                if ((value = clientInfo["client_output_muted"]) != null) if (Boolean.TryParse(value, out bValue)) advOutputMuted = bValue; else advOutputMuted = null; else advOutputMuted = null;
                if ((value = clientInfo["client_outputonly_muted"]) != null) if (Boolean.TryParse(value, out bValue)) advOutputMutedOnly = bValue; else advOutputMutedOnly = null; else advOutputMutedOnly = null;
                if ((value = clientInfo["client_input_hardware"]) != null) if (Boolean.TryParse(value, out bValue)) advInputHardware = bValue; else advInputHardware = null; else advInputHardware = null;
                if ((value = clientInfo["client_output_hardware"]) != null) if (Boolean.TryParse(value, out bValue)) advOutputHardware = bValue; else advOutputHardware = null; else advOutputHardware = null;
                if ((value = clientInfo["client_is_recording"]) != null) if (Boolean.TryParse(value, out bValue)) advIsRecording = bValue; else advIsRecording = null; else advIsRecording = null;

                advFlagAvatar = clientInfo["client_flag_avatar"];
                advAwayMessage = clientInfo["client_away_message"];
                advTalkMessage = clientInfo["client_talke_request_msg"];
                advPhoneticNick = clientInfo["client_nickname_phonetic"];
                advDefaultToken = clientInfo["client_default_token"];
                advBase64Hash = clientInfo["client_base64HashClientUID"];
                if ((value = clientInfo["client_talk_power"]) != null) if (Int32.TryParse(value, out iValue)) advTalkPower = iValue; else advTalkPower = null; else advTalkPower = null;
                if ((value = clientInfo["client_needed_serverquery_view_power"]) != null) if (Int32.TryParse(value, out iValue)) advQueryViewPower = iValue; else advQueryViewPower = null; else advQueryViewPower = null;
                if ((value = clientInfo["client_unread_messages"]) != null) if (Int32.TryParse(value, out iValue)) advUnreadMessages = iValue; else advUnreadMessages = null; else advUnreadMessages = null;
                if ((value = clientInfo["client_icon_id"]) != null) if (Int32.TryParse(value, out iValue)) advIconId = iValue; else advIconId = null; else advIconId = null;
                if ((value = clientInfo["client_away"]) != null) if (Boolean.TryParse(value, out bValue)) advIsAway = bValue; else advIsAway = null; else advIsAway = null;
                if ((value = clientInfo["client_talk_request"]) != null) if (Boolean.TryParse(value, out bValue)) advTalkRequest = bValue; else advTalkRequest = null; else advTalkRequest = null;
                if ((value = clientInfo["client_is_talker"]) != null) if (Boolean.TryParse(value, out bValue)) advIsTalker = bValue; else advIsTalker = null; else advIsTalker = null;
                if ((value = clientInfo["client_is_priority_speaker"]) != null) if (Boolean.TryParse(value, out bValue)) advIsPriority = bValue; else advIsPriority = null; else advIsPriority = null;

                if ((value = clientInfo["client_month_bytes_uploaded"]) != null) if (Int32.TryParse(value, out iValue)) advBytesUpMonth = iValue; else advBytesUpMonth = null; else advBytesUpMonth = null;
                if ((value = clientInfo["client_month_bytes_downloaded"]) != null) if (Int32.TryParse(value, out iValue)) advBytesDownMonth = iValue; else advBytesDownMonth = null; else advBytesDownMonth = null;
                if ((value = clientInfo["client_total_bytes_uploaded"]) != null) if (Int32.TryParse(value, out iValue)) advBytesUpTotal = iValue; else advBytesUpTotal = null; else advBytesUpTotal = null;
                if ((value = clientInfo["client_total_bytes_downloaded"]) != null) if (Int32.TryParse(value, out iValue)) advBytesDownTotal = iValue; else advBytesDownTotal = null; else advBytesDownTotal = null;
                if ((value = clientInfo["connection_filetransfer_bandwidth_sent"]) != null) if (Int32.TryParse(value, out iValue)) advFileBandwidthSent = iValue; else advFileBandwidthSent = null; else advFileBandwidthSent = null;
                if ((value = clientInfo["connection_filetransfer_bandwidth_received"]) != null) if (Int32.TryParse(value, out iValue)) advFileBandwidthRec = iValue; else advFileBandwidthRec = null; else advFileBandwidthRec = null;
                if ((value = clientInfo["connection_packets_sent_total"]) != null) if (Int32.TryParse(value, out iValue)) advPacketsTotalSent = iValue; else advUnreadMessages = null; else advUnreadMessages = null;
                if ((value = clientInfo["connection_packets_received_total"]) != null) if (Int32.TryParse(value, out iValue)) advPacketsTotalSent = iValue; else advPacketsTotalSent = null; else advPacketsTotalSent = null;
                if ((value = clientInfo["connection_bytes_sent_total"]) != null) if (Int32.TryParse(value, out iValue)) advBytesTotalSent = iValue; else advBytesTotalSent = null; else advBytesTotalSent = null;
                if ((value = clientInfo["connection_bytes_received_total"]) != null) if (Int32.TryParse(value, out iValue)) advBytesTotalRec = iValue; else advBytesTotalRec = null; else advBytesTotalRec = null;
                if ((value = clientInfo["connection_bandwidth_sent_last_second_total"]) != null) if (Int32.TryParse(value, out iValue)) advBndwdthSecondSent = iValue; else advBndwdthSecondSent = null; else advBndwdthSecondSent = null;
                if ((value = clientInfo["connection_bandwidth_received_last_second_total"]) != null) if (Int32.TryParse(value, out iValue)) advBndwdthSecondRec = iValue; else advBndwdthSecondRec = null; else advBndwdthSecondRec = null;
                if ((value = clientInfo["connection_bandwidth_sent_last_minute_total"]) != null) if (Int32.TryParse(value, out iValue)) advBndwdthMinuteSent = iValue; else advBndwdthMinuteSent = null; else advBndwdthMinuteSent = null;
                if ((value = clientInfo["connection_bandwidth_received_last_minute_total"]) != null) if (Int32.TryParse(value, out iValue)) advBndwdthMinuteRec = iValue; else advBndwdthMinuteRec = null; else advBndwdthMinuteRec = null;
            }
        }

        #endregion
        #region Helpers

        /// <summary>Contains helper methods to quickly perform commonly used actions.</summary>
        public static class TeamspeakHelper
        {
            /// <summary>Escapes all special characters in a string to conform to TS3 standards.</summary>
            /// <param name="text">The text to escape.</param>
            /// <returns>The escaped text.</returns>
            public static String ts_EscapeString(String text)
            {
                // Replace all special characters with escaped characters.
                String escaped = text.Replace("\\", @"\\");
                escaped = escaped.Replace("/", @"\/");
                escaped = escaped.Replace(" ", @"\s");
                escaped = escaped.Replace("|", @"\p");
                escaped = escaped.Replace("\a", @"\a");
                escaped = escaped.Replace("\b", @"\b");
                escaped = escaped.Replace("\f", @"\f");
                escaped = escaped.Replace("\n", @"\n");
                escaped = escaped.Replace("\r", @"\r");
                escaped = escaped.Replace("\t", @"\t");
                escaped = escaped.Replace("\v", @"\v");
                return escaped;
            }

            /// <summary>Unescapes all special characters in a string from TS3 standards.</summary>
            /// <param name="text">The text to unescape.</param>
            /// <returns>The unescaped text.</returns>
            public static String ts_UnescapeString(String text)
            {
                // Replace all special characters with escaped characters.
                String unescaped = text.Replace(@"\\", "\\");
                unescaped = unescaped.Replace(@"\/", "/");
                unescaped = unescaped.Replace(@"\s", " ");
                unescaped = unescaped.Replace(@"\p", "|");
                unescaped = unescaped.Replace(@"\a", "\a");
                unescaped = unescaped.Replace(@"\b", "\b");
                unescaped = unescaped.Replace(@"\f", "\f");
                unescaped = unescaped.Replace(@"\n", "\n");
                unescaped = unescaped.Replace(@"\r", "\r");
                unescaped = unescaped.Replace(@"\t", "\t");
                unescaped = unescaped.Replace(@"\v", "\v");
                return unescaped;
            }
        }

        #endregion

        #endregion

        #region Plugin Information

        /// <summary>Allows PRoCon to get the name of this plugin.</summary>
        /// <returns>The name of this plugin.</returns>
        public string GetPluginName() { return "Teamspeak 3 Sync"; }

        /// <summary>Allows PRoCon to get the version of this plugin.</summary>
        /// <returns>The version of this plugin.</returns>
        public string GetPluginVersion() { return "0.9.3"; }

        /// <summary>Allows PRoCon to get the author's name of this plugin.</summary>
        /// <returns>The author's name of this plugin.</returns>
        public string GetPluginAuthor() { return "Imisnew2"; }

        /// <summary>Allows PRoCon to get the website for this plugin.</summary>
        /// <returns>The website for this plugin.</returns>
        public string GetPluginWebsite() { return "www.TeamPlayerGaming.com/members/Imisnew2.html"; }

        /// <summary>Allows ProCon to get a description of this plugin.</summary>
        /// <returns>The description of this plugin.</returnsd>
        public string GetPluginDescription()
        {
            return "<h2>Description</h2>" +
                       "<p>Teamspeak 3 Sync synchronizes your Bad Company 2 server with a Teamspeak 3 server, moving players into their respective channels on Teamspeak dependent on the player's team in Bad Company 2." +
                          "<br/>" +
                          "The plugin also allows you to 'yell' at players who are not in your Teamspeak server, yet playing on your Bad Company 2 server.</p>" +

                    "<h2>Setup Instructions</h2>" +

                       "<h3>PRoCon Setup:</h3>" +
                          "<ul style=\"margin-top: 0px;\">" +
                             "<li>Copy the <i>plugin</i> (TeamspeakSync.cs) into your plugins directory (Plugins/BFBC2) <b>OR</b> install the plugin from the packages tab within PRoCon.</li>" +
                             "<li>In your <i>config file</i> (Configs\\procon.cfg), make sure the following settings have the specified values:</li>" +
                                "<ul>" +
                                   "<li>procon.private.options.runPluginsInSandbox <b>True</b></li>" +
                                   "<li>procon.private.options.trustedHostDomainsPorts <b>[ip_of_teamspeak] [query_port_of_teamspeak]</b></li>" +
                                "</ul>" +
                             "<li>Make sure the box PRoCon is being hosted on <i>allows outgoing and incoming connections</i>. This plugin must connect to your Teamspeak 3 server using TCP/IP. It binds to a random local port that is open at the time the plugin is enabled.</li>" +
                          "</ul>" +

                       "<h3>Teamspeak Setup:</h3>" +
                          "<ul style=\"margin-top: 0px;\">" +
                             "<li>Add your PRoCon's IP to query_ip_whitelist.txt. If your Teamspeak 3 server is being hosted by someone else, ask your provider to do this for you..</li>" +
                             "<li>Setup your ServerQuery Login through Teamspeak by going to <b>Tools -> ServerQuery Login -> *Enter A Username* -> *Receive Auto Generated Password*</b>.</li>" +
                             "<li>To find what port your Teamspeak 3 server uses as its <i>query port</i>, look in the server.ini file. The default value is 10011. If your Teamspeak 3 server is being hosted by someone else, ask your provider for the query port.</li>" +
                          "</ul>" +

                     "<h2>Settings</h2>" +
                       "<p>I've tried to let the user control every aspect of the program, while still leaving the interface easy to use.  Below, you'll find descriptions of each of the plugin's various settings.</p>" +

                       "<h3>Connection Error Settings</h3>" +

                         "<h4>Reconnect On Error</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls whether to attempt to re-establish a connection to the Teamspeak server if it encounters a connection error while operating." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Yes</i>: Attempt to reconnect." +
                           "<br/><i>No</i>: Do not attempt to reconnect.</blockquote>" +

                         "<h4>Number of Reconnect Attempts</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the number of attempts when trying to re-establish a connection to the Teamspeak server." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value equal to or above 0.  By specifying 0, the connection will attempt to reconnect indefinitely.</blockquote>" +

                         "<h4>Interval Between Reconnect Attempts (ms)</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the amount of delay (in milliseconds) between attempts to re-establish a connection to the Teamspeak server." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value above 999.</blockquote>" +

                       "<h3>Debug Settings</h3>" +

                         "<h4>Show Debug Messages</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls whether to display debugging information in the plugin console." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Yes</i>: Display debugging information." +
                           "<br/><i>No</i>: Do not display debugging information.</blockquote>" +

                       "<h3>Synchronization Settings</h3>" +

                         "<h4>Matching Threshold (%)</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls how close a player's name must match between the Teamspeak server and Bad Company 2 server." +
                           "<br/>This is only used if there are multiple clients with the same IP, or a client's IP could not be found." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value between 0 and 100, inclusive.</blockquote>" +

                         "<h4>Update Interval (ms)</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the amount of delay (in milliseconds) between queries to the Teamspeak server." +
                           "<br/>Increasing this value will lower the accuracy of the plugin but also lower the load on your bandwidth." +
                           "<br/>Decreasing this value will raise the accuracy of the plugin but also raise the load on your bandwidth." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value above 3999.</blockquote>" +

                         "<h4>Team-Based Threshold</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The number of players that must be present on both your Teamspeak 3 server and Bad Company 2 server before Team Channels are used to divide up players." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical equal to or above -1.  Specifying -1 will turn off Team-Based swapping.  Specifying 0 will turn on Team-Based swapping permanently.</blockquote>" +

                         "<h4>Team Channel Names</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the names given to the team channels.  If left empty, Team 1, Team 2, etc will be used." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>String Array</i>: Multiple channel names.</blockquote>" +

                         "<h4>Squad-Based Threshold</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The number of players that must be present on both your Teamspeak 3 server and Bad Company 2 server before Squad Channels are used to divide up players." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical equal to or above -1.  Specifying -1 will turn off Squad-Based swapping.  Specifying 0 will turn on Squad-Based swapping permanently.</blockquote>" +

                         "<h4>Squad Channel Names</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the names given to the squad channels.  If left empty, Squad 1, Squad 2, etc will be used." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>String Array</i>: Multiple channel names.</blockquote>" +

                         "<h4>Remove Channels When Empty</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls whether to remove channels as they become empty." +
                           "<br/>Creating and deleting channels is <i>very</i> time consuming.  Enabling this feature may cause swapping to become sluggish with a small amount of players." +
                           "<br/>Enabling this feature will also create, relatively, a large bandwidth increase." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Yes</i>: Remove channels as they become empty." +
                           "<br/><i>No</i>: Do not remove channels ever.</blockquote>" +

                         "<h4>Remove Clients Not Playing</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls whether to remove clients from the Team and Squad channels if they are not playing on your Bad Company 2 server." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Yes</i>: Remove clients from the Team and Squad channels." +
                           "<br/><i>No</i>: Do not remove clients from the Team and Squad channels.</blockquote>" +

                         "<h4>Remove Clients - Whitelist</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">When removing clients from the Team and Squad channels, clients who are listed here will not be removed." +
                           "<br/>The client's Database Id is used to determine whether they should be removed or not." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value that represents a client's Database Id.  When specifying multiple values, separate them with '|'.  E.g. '15|32|412|22'</blockquote>" +

                       "<h3>Teamspeak 3 Server</h3>" +

                         "<h4>Ts3 Server IP</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The IP of the Teamspeak 3 Server you want to synchronize to your Bad Company 2 server." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Text</i>: An IP, such as 127.0.0.1, or a Host Name, such as ts3.myteamspeakserver.com.</blockquote>" +

                         "<h4>Ts3 Server Port</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The Port Number that a regular client would use to connect to your Teamspeak 3 server." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value between 0 and 65535, inclusive.</blockquote>" +

                         "<h4>Ts3 Query Port</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The Port Number that a Server Query Client would use to login to your Teamspeak 3 server." +
                           "<br/>If you do not know this value and your server is being hosted by someone other than yourself, ask your server provider what the Server Query Port Number is." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value between 0 and 65535, inclusive.</blockquote>" +

                         "<h4>Ts3 Query Username</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The username of your Server Query Login." +
                           "<br/>To setup this value, open up and connect to your Teamspeak server.  Go to Tools -> ServerQuery Login and specify what you want your username to be." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Text</i>: A username.</blockquote>" +

                         "<h4>Ts3 Query Password</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The password of your Server Query Login." +
                           "<br/>See Ts3 Server Query Username on how to setup this value." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Text</i>: A password.</blockquote>" +

                         "<h4>Staging Channel Name</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">The Name of the Channel users will join to be swapped to the correct Team Channels." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Text</i>: The Name of the Staging Channel as it appears in Teamspeak.</blockquote>" +

                       "<h3>Yell Settings</h3>" +

                         "<h4>Yell To Players Not In Teamspeak</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls whether players who are not in Teamspeak will be yelled to." +
                           "<br/>Turning this on effectively turns on \"Yelling In Intervals\"." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Yes</i>: Yell to players who are not connected to your Teamspeak server." +
                           "<br/><i>No</i>: Do not yell to players who are not connect to your Teamspeak server.</blockquote>" +

                         "<h4>Yell When Player Joins</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls whether players are yelled to right when they join the server, if they are not in your Teamspeak server." +
                           "<br/>Exactly when the player is yelled to upon joining the game is controlled by the next setting." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Yes</i>: Yell to players who are not connected to your Teamspeak server when they join your Bad Company 2 server." +
                           "<br/><i>No</i>: Do not yell to players who are not connected to your Teamspeak server when they join your Bad Company 2 server.</blockquote>" +

                         "<h4>Yell When Player Joins Delay (ms)</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the amount of delay (in milliseconds) between when a player joins your Bad Company 2 server and when they are yelled to." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value greater than or equal to 0.</blockquote>" +

                         "<h4>Yell Interval (ms)</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the amount of delay (in milliseconds) between yells to all players who are not in your Teamspeak server." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value greater than or equal to the yell message's duration.</blockquote>" +

                         "<h4>Yell Message</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls what is displayed when a yell is sent to a player." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Text</i>: The message to send to the player.</blockquote>" +

                         "<h4>Yell Duration (ms)</h4>" +
                         "<blockquote style=\"margin-left: 0px; margin-right:0px; margin-top:0px;\">Controls the time (in milliseconds) the yell is displayed on the player's screen." +
                           "<br/>" +
                           "<br/><u>Values</u>:" +
                           "<br/><i>Number</i>: Some numerical value above 999.</blockquote>";
        }

        #endregion
        #region Plugin PRoCon Variables

        #region UI Variables

        // Teamspeak Server Settings
        String uiTs3ServerIp = "Teamspeak Ip";
        UInt16 uiTs3ServerPort = 9987;
        UInt16 uiTs3QueryPort = 10011;
        String uiTs3QueryUsername = "Username";
        String uiTs3QueryPassword = "Password";
        String uiTs3StgChannelName = "Staging Channel Name";

        // Synchronization Settings
        Double uiMatchingThreshold = 75.0;
        Int32 uiUpdateInterval = 10000;
        Int32 uiTeamBasedThreshold = 0;
        String[] uiTeamChannelNames = new String[] { "Team 1", "Team 2", "Team 3", "Team 4" };
        Int32 uiSquadBasedThreshold = 8;
        String[] uiSquadChannelNames = new String[] { "Alpha", "Bravo", "Charlie", "Delta", "Echo", "Foxtrot", "Golf", "Hotel" };
        enumBoolYesNo uiRemoveChannelsOnEmpty = enumBoolYesNo.No;
        enumBoolYesNo uiRemoveClients = enumBoolYesNo.Yes;
        String uiRemoveClientsWhitelist = "";

        // Yell Settings
        enumBoolYesNo uiYelling = enumBoolYesNo.Yes;
        enumBoolYesNo uiYellOnJoin = enumBoolYesNo.No;
        Int32 uiYellOnJoinDelay = 300000;
        Int32 uiYellingInterval = 1800000;
        String uiYellMessage = "Yell Message";
        Int32 uiYellDuration = 6000;

        // Connection Error Settings
        enumBoolYesNo uiReconnectOnError = enumBoolYesNo.Yes;
        Int32 uiReconnectOnErrorAttempts = 10;
        Int32 uiReconnectOnErrorInterval = 5000;

        // Debug Settings
        enumBoolYesNo uiDebug = enumBoolYesNo.No;

        #endregion

        /// <summary>Allows PRoCon to figure out what fields to display to the user.</summary>
        /// <returns>A list of variables to display.</returns>
        public List<CPluginVariable> GetDisplayPluginVariables()
        {
            List<CPluginVariable> lstReturn = new List<CPluginVariable>();

            // Teamspeak Server Settings
            lstReturn.Add(new CPluginVariable("Teamspeak 3 Server|Ts3 Server IP", typeof(String), uiTs3ServerIp));
            lstReturn.Add(new CPluginVariable("Teamspeak 3 Server|Ts3 Server Port", typeof(Int32), uiTs3ServerPort));
            lstReturn.Add(new CPluginVariable("Teamspeak 3 Server|Ts3 Query Port", typeof(Int32), uiTs3QueryPort));
            lstReturn.Add(new CPluginVariable("Teamspeak 3 Server|Ts3 Query Username", typeof(String), uiTs3QueryUsername));
            lstReturn.Add(new CPluginVariable("Teamspeak 3 Server|Ts3 Query Password", typeof(String), uiTs3QueryPassword));
            lstReturn.Add(new CPluginVariable("Teamspeak 3 Server|Staging Channel Name", typeof(String), uiTs3StgChannelName));

            // Synchronization Settings
            lstReturn.Add(new CPluginVariable("Synchronization|Matching Threshold (%)", typeof(Double), uiMatchingThreshold));
            lstReturn.Add(new CPluginVariable("Synchronization|Update Interval (ms)", typeof(Int32), uiUpdateInterval));
            lstReturn.Add(new CPluginVariable("Synchronization|Team-Based Threshold", typeof(Int32), uiTeamBasedThreshold));
            lstReturn.Add(new CPluginVariable("Synchronization|Team Channel Names", typeof(String[]), uiTeamChannelNames));
            if (uiTeamBasedThreshold != -1)
            {
                lstReturn.Add(new CPluginVariable("Synchronization|Squad-Based Threshold", typeof(Int32), uiSquadBasedThreshold));
                lstReturn.Add(new CPluginVariable("Synchronization|Squad Channel Names", typeof(String[]), uiSquadChannelNames));
            }
            lstReturn.Add(new CPluginVariable("Synchronization|Remove Channels When Empty", typeof(enumBoolYesNo), uiRemoveChannelsOnEmpty));
            lstReturn.Add(new CPluginVariable("Synchronization|Remove Clients Not Playing", typeof(enumBoolYesNo), uiRemoveClients));
            if (uiRemoveClients == enumBoolYesNo.Yes)
                lstReturn.Add(new CPluginVariable("Synchronization|Remove Clients - Whitelist", typeof(String), uiRemoveClientsWhitelist));

            // Yell Settings
            lstReturn.Add(new CPluginVariable("Yell Settings|Yell To Players Not In Teamspeak", typeof(enumBoolYesNo), uiYelling));
            if (uiYelling == enumBoolYesNo.Yes)
            {
                lstReturn.Add(new CPluginVariable("Yell Settings|Yell When Player Joins", typeof(enumBoolYesNo), uiYellOnJoin));
                if (uiYellOnJoin == enumBoolYesNo.Yes)
                    lstReturn.Add(new CPluginVariable("Yell Settings|Yell When Player Joins Delay (ms)", typeof(Int32), uiYellOnJoinDelay));
                lstReturn.Add(new CPluginVariable("Yell Settings|Yell Interval (ms)", typeof(Int32), uiYellingInterval));
                lstReturn.Add(new CPluginVariable("Yell Settings|Yell Message", typeof(String), uiYellMessage));
                lstReturn.Add(new CPluginVariable("Yell Settings|Yell Display Duration (ms)", typeof(Int32), uiYellDuration));
            }

            // Connection Error Settings
            lstReturn.Add(new CPluginVariable("Connection Error Settings|Reconnect On Error", typeof(enumBoolYesNo), uiReconnectOnError));
            if (uiReconnectOnError == enumBoolYesNo.Yes)
            {
                lstReturn.Add(new CPluginVariable("Connection Error Settings|Number of Reconnect Attempts", typeof(Int32), uiReconnectOnErrorAttempts));
                lstReturn.Add(new CPluginVariable("Connection Error Settings|Interval Between Reconnect Attempts (ms)", typeof(Int32), uiReconnectOnErrorInterval));
            }

            // Debug Settings
            lstReturn.Add(new CPluginVariable("Debug|Show Debug Messages", typeof(enumBoolYesNo), uiDebug));

            return lstReturn;
        }

        /// <summary>Allows PRoCon to save the variables for persistence across sessions.</summary>
        /// <returns>A list of variables to save.</returns>
        public List<CPluginVariable> GetPluginVariables()
        {
            List<CPluginVariable> lstReturn = new List<CPluginVariable>();

            // Teamspeak Server Settings
            lstReturn.Add(new CPluginVariable("Ts3 Server IP", typeof(String), uiTs3ServerIp));
            lstReturn.Add(new CPluginVariable("Ts3 Server Port", typeof(Int32), uiTs3ServerPort));
            lstReturn.Add(new CPluginVariable("Ts3 Query Port", typeof(Int32), uiTs3QueryPort));
            lstReturn.Add(new CPluginVariable("Ts3 Query Username", typeof(String), uiTs3QueryUsername));
            lstReturn.Add(new CPluginVariable("Ts3 Query Password", typeof(String), uiTs3QueryPassword));
            lstReturn.Add(new CPluginVariable("Staging Channel Name", typeof(String), uiTs3StgChannelName));

            // Synchronization Settings
            lstReturn.Add(new CPluginVariable("Matching Threshold (%)", typeof(Double), uiMatchingThreshold));
            lstReturn.Add(new CPluginVariable("Update Interval (ms)", typeof(Int32), uiUpdateInterval));
            lstReturn.Add(new CPluginVariable("Team-Based Threshold", typeof(Int32), uiTeamBasedThreshold));
            lstReturn.Add(new CPluginVariable("Team Channel Names", typeof(String[]), uiTeamChannelNames));
            lstReturn.Add(new CPluginVariable("Squad-Based Threshold", typeof(Int32), uiSquadBasedThreshold));
            lstReturn.Add(new CPluginVariable("Squad Channel Names", typeof(String[]), uiSquadChannelNames));
            lstReturn.Add(new CPluginVariable("Remove Channels When Empty", typeof(enumBoolYesNo), uiRemoveChannelsOnEmpty));
            lstReturn.Add(new CPluginVariable("Remove Clients Not Playing", typeof(enumBoolYesNo), uiRemoveClients));
            lstReturn.Add(new CPluginVariable("Remove Clients - Whitelist", typeof(String), uiRemoveClientsWhitelist));

            // Yell Settings
            lstReturn.Add(new CPluginVariable("Yell To Players Not In Teamspeak", typeof(enumBoolYesNo), uiYelling));
            lstReturn.Add(new CPluginVariable("Yell When Player Joins", typeof(enumBoolYesNo), uiYellOnJoin));
            lstReturn.Add(new CPluginVariable("Yell When Player Joins Delay (ms)", typeof(Int32), uiYellOnJoinDelay));
            lstReturn.Add(new CPluginVariable("Yell Interval (ms)", typeof(Int32), uiYellingInterval));
            lstReturn.Add(new CPluginVariable("Yell Message", typeof(String), uiYellMessage));
            lstReturn.Add(new CPluginVariable("Yell Display Duration (ms)", typeof(Int32), uiYellDuration));

            // Connection Error Settings
            lstReturn.Add(new CPluginVariable("Reconnect On Error", typeof(enumBoolYesNo), uiReconnectOnError));
            lstReturn.Add(new CPluginVariable("Number of Reconnect Attempts", typeof(Int32), uiReconnectOnErrorAttempts));
            lstReturn.Add(new CPluginVariable("Interval Between Reconnect Attempts (ms)", typeof(Int32), uiReconnectOnErrorInterval));

            // Debug Settings
            lstReturn.Add(new CPluginVariable("Show Debug Messages", typeof(enumBoolYesNo), uiDebug));

            return lstReturn;
        }

        /// <summary>Allows PRoCon to load the variables from a previous session.  Also allows the user to change the variable as well.</summary>
        /// <param name="strVariable">The name of the variable we're loading.</param>
        /// <param name="strValue">The value of the variable we're loading.</param>
        public void SetPluginVariable(string strVariable, string strValue)
        {
            //Temporary out variable for TryParse methods.
            UInt16 ushOut = 0;
            Int32 intOut = 0;
            Double dblOut = 0;

            #region Teamspeak Server Settings

            if (strVariable == "Ts3 Server IP")
                ts3Connection.IP = uiTs3ServerIp = strValue;
            else if (strVariable == "Ts3 Server Port" && UInt16.TryParse(strValue, out ushOut))
                ts3Port = uiTs3ServerPort = ushOut;
            else if (strVariable == "Ts3 Query Port" && UInt16.TryParse(strValue, out ushOut))
                ts3Connection.Port = uiTs3QueryPort = ushOut;
            else if (strVariable == "Ts3 Query Username")
                ts3Username = uiTs3QueryUsername = strValue;
            else if (strVariable == "Ts3 Query Password")
                ts3Password = uiTs3QueryPassword = strValue;
            else if (strVariable == "Staging Channel Name")
                ts3StagingChannel.tsName = uiTs3StgChannelName = strValue;

            #endregion
            #region Synchronization Settings

            else if (strVariable == "Matching Threshold (%)" && Double.TryParse(strValue, out dblOut))
                synMatchingThreshold = uiMatchingThreshold = (dblOut >= 0.0 && dblOut <= 100.0) ? dblOut : 75.0;
            else if (strVariable == "Update Interval (ms)" && Int32.TryParse(strValue, out intOut))
                synUpdateInterval = uiUpdateInterval = (intOut >= 4000) ? intOut : 10000;
            else if (strVariable == "Team-Based Threshold" && Int32.TryParse(strValue, out intOut))
                synTeamBasedThreshold = uiTeamBasedThreshold = (intOut >= -1 && intOut <= 32) ? intOut : 0;
            else if (strVariable == "Team Channel Names")
                synTeamChannelNames = uiTeamChannelNames = CPluginVariable.DecodeStringArray(strValue);
            else if (strVariable == "Squad-Based Threshold" && Int32.TryParse(strValue, out intOut))
                synSquadBasedThreshold = uiSquadBasedThreshold = (intOut >= -1 && intOut <= 32) ? intOut : 8;
            else if (strVariable == "Squad Channel Names")
                synSquadChannelNames = uiSquadChannelNames = CPluginVariable.DecodeStringArray(strValue);
            else if (strVariable == "Remove Channels When Empty" && Enum.IsDefined(typeof(enumBoolYesNo), strValue))
                synRemoveChannelsOnEmpty = (uiRemoveChannelsOnEmpty = (enumBoolYesNo)Enum.Parse(typeof(enumBoolYesNo), strValue)) == enumBoolYesNo.Yes;
            else if (strVariable == "Remove Clients Not Playing" && Enum.IsDefined(typeof(enumBoolYesNo), strValue))
                synRemoveClients = (uiRemoveClients = (enumBoolYesNo)Enum.Parse(typeof(enumBoolYesNo), strValue)) == enumBoolYesNo.Yes;
            else if (strVariable == "Remove Clients - Whitelist")
            {
                uiRemoveClientsWhitelist = strValue;

                // Update whitelist.
                synRemoveClientsWhitelist.Clear();
                foreach (String id in uiRemoveClientsWhitelist.Split('|'))
                    if (Int32.TryParse(id.Trim(), out intOut))
                        synRemoveClientsWhitelist.Add(intOut);
            }

            #endregion
            #region Yell Settings

            else if (strVariable == "Yell To Players Not In Teamspeak" && Enum.IsDefined(typeof(enumBoolYesNo), strValue))
                yllYelling = (uiYelling = (enumBoolYesNo)Enum.Parse(typeof(enumBoolYesNo), strValue)) == enumBoolYesNo.Yes;
            else if (strVariable == "Yell When Player Joins" && Enum.IsDefined(typeof(enumBoolYesNo), strValue))
                yllOnJoin = (uiYellOnJoin = (enumBoolYesNo)Enum.Parse(typeof(enumBoolYesNo), strValue)) == enumBoolYesNo.Yes;
            else if (strVariable == "Yell When Player Joins Delay (ms)" && Int32.TryParse(strValue, out intOut))
                yllOnJoinDelay = uiYellOnJoinDelay = (intOut >= 0) ? intOut : 300000;
            else if (strVariable == "Yell Interval (ms)" && Int32.TryParse(strValue, out intOut))
                yllInterval = uiYellingInterval = (intOut >= yllDuration) ? intOut : yllDuration;
            else if (strVariable == "Yell Message")
                yllMessage = uiYellMessage = strValue;
            else if (strVariable == "Yell Display Duration (ms)" && Int32.TryParse(strValue, out intOut))
            {
                yllDuration = uiYellDuration = (intOut >= 1000) ? intOut : 1000;
                yllInterval = uiYellingInterval = (uiYellDuration > uiYellingInterval) ? uiYellDuration : uiYellingInterval;
            }

            #endregion
            #region Connection Error Settings

            else if (strVariable == "Reconnect On Error" && Enum.IsDefined(typeof(enumBoolYesNo), strValue))
                roeReconnectOnError = (uiReconnectOnError = (enumBoolYesNo)Enum.Parse(typeof(enumBoolYesNo), strValue)) == enumBoolYesNo.Yes;
            else if (strVariable == "Number of Reconnect Attempts" && int.TryParse(strValue, out intOut))
                roeReconnectOnErrorAttempts = uiReconnectOnErrorAttempts = (intOut >= 0) ? intOut : 10;
            else if (strVariable == "Interval Between Reconnect Attempts (ms)" && int.TryParse(strValue, out intOut))
                roeReconnectOnErrorInterval = uiReconnectOnErrorInterval = (intOut >= 1000) ? intOut : 15000;

            #endregion
            #region Debug Settings

            else if (strVariable == "Show Debug Messages" && Enum.IsDefined(typeof(enumBoolYesNo), strValue))
                mstDebug = (uiDebug = (enumBoolYesNo)Enum.Parse(typeof(enumBoolYesNo), strValue)) == enumBoolYesNo.Yes;

            #endregion
        }

        #endregion
        #region Plugin Variables

        #region ----------- TEAMSPEAK 3 VARIABLES -----------
        // -- Visual settings
        // [0] The connection to the Teamspeak server.
        // [1] The port for regular clients on the Teamspeak server.
        // [2] The ServerQuery username.
        // [3] The ServerQuery password.
        // [4] The Staging Channel's information.
        TeamspeakConnection ts3Connection;
        Int32 ts3Port;
        String ts3Username;
        String ts3Password;
        TeamspeakChannel ts3StagingChannel;
        // -- Behind the scenes settings
        // [0] The Team Channels' information (Team ID, Team Channel's Info).
        // [1] The Squad Channels' information (Team ID, Squad Channels' Info).
        // [2] The clients-who-are-on-teamspeak's info.
        // [3] The response (used to hold err. message) from the server.
        Dictionary<Int32, TeamspeakChannel> ts3TeamChannels;
        Dictionary<Int32, Dictionary<Int32, TeamspeakChannel>> ts3SquadChannels;
        List<TeamspeakClient> ts3ClientInfo;
        TeamspeakResponse ts3Response;
        #endregion

        #region ----------- SYNCHRONIZATION VARIABLES -----------
        // -- Visual settings
        // [0] The percent to which player's names have to match to be considered the same player.
        // [1] The amount of milliseconds between synchronization efforts.
        // [2] The amount of people that must be on teamspeak before clients are moved to a team-based channel.
        // [3] The amount of people that must be on a specific team in teamspeak before clients are moved to a squad-based channels.
        // [4] Whether to remove clients in from att/def channels who aren't playing in bad company 2.
        // [5] The list of client's database ids to not remove from att/def channels, if previous true.
        Double synMatchingThreshold;
        Int32 synUpdateInterval;
        Int32 synTeamBasedThreshold;
        String[] synTeamChannelNames;
        Int32 synSquadBasedThreshold;
        String[] synSquadChannelNames;
        Boolean synRemoveChannelsOnEmpty;
        Boolean synRemoveClients;
        List<Int32> synRemoveClientsWhitelist;
        // -- Behind the scenes settings
        // [0] The thread used to handle updating.
        Thread synUpdateThread;
        #endregion

        #region ----------- YELL VARIABLES -----------
        // -- Visual settings
        // [0] Whether to yell to players at set intervals.
        // [1] The delay, in milliseconds, between intervals.
        // [2] Whether to yell to the players when the join the server (in addition to intervals).
        // [3] The delay, in milliseconds, between when a player joins the server and when he/she is yelled to.
        // [4] The message to yell.
        // [5] The duration, in milliseconds, to display the yell.
        Boolean yllYelling;
        Int32 yllInterval;
        Boolean yllOnJoin;
        Int32 yllOnJoinDelay;
        String yllMessage;
        Int32 yllDuration;
        // -- Behind the scenes settings
        // [0] The thread used to handle yelling in intervals.
        // [1] The timers used to handle yelling to players as they join the server.
        Thread yllYellingThread;
        Queue<Timer> yllOnJoinTimers;
        #endregion

        #region ----------- RECONNECT ON ERROR VARIABLES -----------
        // -- Visual settings
        // [0] Whether to try and attempt to reconnect to the teamspeak server after a connection error.
        // [1] The number of times a reconnect should be attempted after a connection error.
        // [2] The delay between reconnect attempts.
        Boolean roeReconnectOnError;
        Int32 roeReconnectOnErrorAttempts;
        Int32 roeReconnectOnErrorInterval;
        // -- Behind the scenes settings
        #endregion

        #region ----------- BAD COMPANY 2 VARIABLES -----------
        // [0] The general/punkbuster client combination.
        // [1] Used to manage the punkbuster listings received.
        List<BadCompanyClient> bc2ClientInfo;
        Dictionary<String, CPunkbusterInfo> bc2PunkbusterInfo;
        #endregion

        #region ----------- UNIVERSAL VARIABLES -----------
        // [0] The paired up clients from teamspeak and bad company 2.
        // [1] The queue of actions to execute.
        // [2] The state of the action queue.
        // [3] Whether to print debug statements.
        List<MasterClient> mstClientInfo;
        Queue<ActionEvent> mstActions;
        Boolean mstRunning;
        Boolean mstDebug;
        // [0] The thread used to execute actions.
        // [1] The event that is fired when an action is added to the queue.
        // [2] Counts how many actions are available.
        // [3] Guards the queue from multiple threads.
        Thread mstActionThread;
        AutoResetEvent mstActionAdded;
        Semaphore mstActionSemaphore;
        Mutex mstActionMutex;
        // [0] Whether the last event was a ON_PLAYER_LEAV event.
        // [1] The name of the last player that left the BC2 server.
        Boolean mstPlayerLeft;
        String mstPlayerName;
        // -- Various Constants
        const Int32 OBTAIN_TS_INFO = 0;
        const Int32 OBTAIN_BC_INFO = 1;
        const Int32 MANAGE_PB_INFO = 2;
        const Int32 UPDATE_MC_INFO = 3;
        const Int32 CHECK_FOR_SWAP = 10;
        const Int32 CHECK_FOR_REMV = 11;
        const Int32 CHECK_FOR_YELL = 12;
        const Int32 CHECK_SWAP_ALL = 13;
        const Int32 CHECK_REMV_ALL = 14;
        const Int32 CHECK_YELL_ALL = 15;
        const Int32 CHECK_YELL_JIN = 16;
        const Int32 ON_PLAYER_JOIN = 20;
        const Int32 ON_PLAYER_LEAV = 21;
        const Int32 ON_PLAYER_TMSQ = 23;
        #endregion

        /// <summary>Holds the player/punkbuster info combo from bad company 2.</summary>
        public class BadCompanyClient
        {
            // [0] The player's general info (Teams/Squads).
            // [1] The player's punkbuster info (IPs).
            private CPlayerInfo generalInfo = null;
            private CPunkbusterInfo punkbusterInfo = null;

            // Easy access to general info.
            public String Name { get { return generalInfo.SoldierName; } }
            public String Tags { get { return generalInfo.ClanTag; } }
            public Int32 TeamId { get { return generalInfo.TeamID; } }
            public Int32 SquadId { get { return generalInfo.SquadID; } }
            public Int32 Score { get { return generalInfo.Score; } }
            public Int32 Kills { get { return generalInfo.Kills; } }
            public Int32 Deaths { get { return generalInfo.Deaths; } }
            public Double KDR { get { return generalInfo.Kdr; } }
            public Int32 Ping { get { return generalInfo.Ping; } }
            public String GUID { get { return generalInfo.GUID; } }
            // Easy access to punkbuster info.
            public String IP { get { if (punkbusterInfo != null) return punkbusterInfo.Ip; else return null; } }
            public String SlotId { get { if (punkbusterInfo != null) return punkbusterInfo.SlotID; else return null; } }
            public String Country { get { if (punkbusterInfo != null) return punkbusterInfo.PlayerCountry; else return null; } }
            public String CountryCode { get { if (punkbusterInfo != null) return punkbusterInfo.PlayerCountryCode; else return null; } }
            // Some extra helpers.
            /// <summary>Whether this client has punkbuster info initialized.</summary>
            public Boolean HasPbInfo { get { return punkbusterInfo != null; } }
            /// <summary>Sets the general information.</summary>
            public CPlayerInfo GeneralInfo { get { return generalInfo; } set { if (value != null) generalInfo = value; } }
            /// <summary>Sets the punkbuster information.</summary>
            public CPunkbusterInfo PunkbusterInfo { get { return punkbusterInfo; } set { if (value != null) punkbusterInfo = value; } }

            /// <summary>Creates the BadCompanyClient without a punkbuster info.</summary>
            /// <param name="genInfo">The player's general information.</param>
            public BadCompanyClient(CPlayerInfo genInfo) { generalInfo = genInfo; }
            /// <summary>Creates the BadCompanyClient with everything.</summary>
            /// <param name="genInfo">The player's general information.</param>
            /// <param name="pbInfo">The player's punkbuster information.</param>
            public BadCompanyClient(CPlayerInfo genInfo, CPunkbusterInfo pbInfo) { generalInfo = genInfo; punkbusterInfo = pbInfo; }
        }

        /// <summary>Holds the combination of TS Client and BC Client information.</summary>
        public class MasterClient
        {
            // [0] The client's teamspeak information.
            // [1] The client's bad company 2 information.
            private TeamspeakClient tsClient = null;
            private BadCompanyClient bcClient = null;

            /// <summary>Whether the master client has teamspeak information.</summary>
            public Boolean HasTsClient { get { return tsClient != null; } }
            /// <summary>Whether the master client has bad company information.</summary>
            public Boolean HasBcClient { get { return bcClient != null; } }
            /// <summary>Sets the teamspeak information.</summary>
            public TeamspeakClient TsClient { get { return tsClient; } set { tsClient = value; } }
            /// <summary>Sets the bad company information</summary>
            public BadCompanyClient BcClient { get { return bcClient; } set { bcClient = value; } }

            /// <summary>Creates the MasterClient with teamspeak information.</summary>
            /// <param name="ts">The teamspeak information.</param>
            public MasterClient(TeamspeakClient ts) { tsClient = ts; }
            /// <summary>Creates the MasterClient with bad company information.</summary>
            /// <param name="bc">The bad company information.</param>
            public MasterClient(BadCompanyClient bc) { bcClient = bc; }

            /// <summary>Calculates the percent match of a substring in another string. The shortest string is used as the substring.</summary>
            /// <param name="s1">The first string.</param>
            /// <param name="s2">The second string.</param>
            /// <returns>The percent match.</returns>
            public static Double calcPercentMatch(String s1, String s2)
            {
                // Variables
                double max;
                double min;
                int levDist;

                // Always use the longest string first (for Levinshtein Distance).
                if (s1.Length >= s2.Length)
                {
                    max = s1.Length;
                    min = s2.Length;
                    levDist = calcLevenshteinDistance(s1, s2);
                }
                else
                {
                    max = s2.Length;
                    min = s1.Length;
                    levDist = calcLevenshteinDistance(s2, s1);
                }

                double percent = (max - levDist) / max;
                double maxPossMatch = min / max;

                // Calc percent to which the string matched.
                // Calc largest possible match (i.e difference in length, if larger string contained exact substring match).
                // Use relative match.
                // Example:
                //  Bob -> B0bWuzHere
                //  Percent match:    (10 - 8) / 10 = 20%
                //  Max poss match:   3 / 10        = 30%
                //  Relative percent: 20% / 30%     = 66.6%
                return (percent / maxPossMatch) * 100;

                // Another Algorithm works as follows:
                // Calc percent to which the string matched.
                // Calc largest possible match (i.e difference in length, if larger string contained exact substring match).
                // Assume 100% match, then remove the relative amount the strings didn't match.
                // Example:
                //  Bob -> B0bWuzHere
                //  Percent match:  (10 - 8) / 10      = 20%
                //  Max poss match: 3 / 10             = 30%
                //  Assume 100%:    100% - (30% - 20%) = 90%
                //return 100 - (maxPossMatch - percent) * 100;
            }
            /// <summary>Returns the number of changes needed to be made to the first string to turn it into the second string.</summary>
            /// <param name="s1">The first string.  Note: This string needs to be the longer of the two.</param>
            /// <param name="s2">The second string.  Note: This string needs to be the shorter of the two.</param>
            /// <returns>The number of changes needed to be made to the first string.</returns>
            public static Int32 calcLevenshteinDistance(String s1, String s2)
            {
                // [0] Length of S1.
                // [1] Length of S2.
                // [2] Where we're at in S1.
                // [3] Where we're at in S2.
                // [4] The Character at posS1 in S1.
                // [5] The Character at posS2 in S2.
                int lenS1, lenS2;
                int posS1, posS2;
                String chrS1, chrS2;

                // Exit early if one is an empty string.
                lenS1 = s1.Length;
                lenS2 = s2.Length;
                if (lenS1 == 0)
                    return lenS2;
                if (lenS2 == 0)
                    return lenS1;

                // Build a matrix.
                // Fill first row with 0,1,2,3,4,5...
                // Fill first column with 0,1,2,3,4,5...
                int[][] matrix = new int[lenS1][];
                for (int i = 0; i < lenS1; i++)
                    matrix[i] = new int[lenS2];

                for (int i = 0; i < lenS1; i++)
                    matrix[i][0] = i;

                for (int i = 0; i < lenS2; i++)
                    matrix[0][i] = i;


                // Start filling the matrix.
                // For each character in the first string...
                for (posS1 = 1; posS1 < lenS1; posS1++)
                {
                    chrS1 = s1.Substring(posS1, 1);

                    // For each character in the second string...
                    for (posS2 = 1; posS2 < lenS2; posS2++)
                    {
                        chrS2 = s2.Substring(posS2, 1);

                        // Determine if they are equal.
                        int cost = 0;
                        if (chrS1 != chrS2)
                            cost = 1;

                        // Calculate which value should be put in this spot of the matrix.
                        int val1 = matrix[posS1 - 1][posS2] + 1;
                        int val2 = matrix[posS1][posS2 - 1] + 1;
                        int val3 = matrix[posS1 - 1][posS2 - 1] + cost;
                        matrix[posS1][posS2] = Math.Min(Math.Min(val1, val2), val3);
                    }
                }

                // Return the number of additions/removals/changes it would take for the first string to match the second string.
                return matrix[lenS1 - 1][lenS2 - 1];
            }
        }

        /// <summary>Holds an action to perform along with it's arguments.</summary>
        public class ActionEvent
        {
            private Int32 command = 0;
            private Queue<Object> arguments = new Queue<Object>();

            public Int32 Command { get { return command; } }
            public Object Argument { get { return arguments.Dequeue(); } }

            /// <summary>Takes all the parameters for an event and creates them as neccesary.</summary>
            /// <param name="command">The command Id.</param>
            /// <param name="arguments">The arguments.</param>
            public ActionEvent(Int32 command, Object[] arguments)
            {
                this.command = command;
                foreach (Object arg in arguments)
                    this.arguments.Enqueue(arg);
            }
        }

        #endregion
        #region Plugin Loaded/Enable/Disable

        /// <summary>Is called when the plugin is successfully loaded.</summary>
        public void OnPluginLoaded(String strHostName, String strPort, String strPRoConVersion)
        {
            // Register Events.
            this.RegisterEvents("OnPlayerJoin", "OnPlayerLeft", "OnPlayerSquadChange", "OnListPlayers", "OnPunkbusterPlayerInfo");
        }

        /// <summary>Starts the action queue.</summary>
        public void OnPluginEnable()
        {
            try
            {
                // Wait until the action queue is done being processed (i.e. someone disabled and re-enabled quickly).
                if (mstActionThread.IsAlive)
                {
                    consoleWrite("^8Actions are still being performed, please wait until synchronization is haulted to re-enable the plugin.");
                    setPluginState(false);
                    return;
                }

                // Clear out all old data.
                performCleaning();

                // Start the action thread.
                mstRunning = true;
                mstActionThread = new Thread(performActions);
                mstActionThread.Start();
            }
            catch (Exception e)
            {
                consoleWrite("^8Fatal Error!");
                consoleWrite("^8^bMessage:^n^0 " + e.Message);
                consoleWrite("^8^bStack Trace:^n^0 " + e.StackTrace);
                setPluginState(false);
            }
        }

        /// <summary>Stops the action queue.</summary>
        public void OnPluginDisable()
        {
            // Stop the action thread.
            mstRunning = false;
            mstActionSemaphore.Release();
            consoleWrite("Stopping Synchronization...");
        }

        #endregion
        #region Plugin Methods

        public void setPluginState(Boolean state)
        { this.ExecuteCommand("procon.protected.plugins.enable", "TeamspeakSync", state.ToString()); }
        public void consoleWrite(String message)
        { this.ExecuteCommand("procon.protected.pluginconsole.write", "Ts3 Sync: " + message); }
        public void debugWrite(String message)
        { if (mstDebug) this.ExecuteCommand("procon.protected.pluginconsole.write", "^7Ts3 Sync - Debug: " + message); }
        public void yellToPlayer(String message, Int32 duration, String player)
        { this.ExecuteCommand("procon.protected.send", "admin.yell", message, duration.ToString(), "player", player); }
        public void forcePbListing()
        { this.ExecuteCommand("procon.protected.send", "punkBuster.pb_sv_command", "pb_sv_plist"); }
        public void forceBcListing()
        { this.ExecuteCommand("procon.protected.send", "admin.listPlayers", "all"); }

        #endregion
        #region Plugin Events

        /// <summary>Is called when a player joins the server.</summary>
        /// <param name="strSoldierName">The player's name.</param>
        public void OnPlayerJoin(String strSoldierName)
        {
            if (mstRunning)
                addToActionQueue(ON_PLAYER_JOIN, strSoldierName);
        }

        /// <summary>Is called when a player leaves the server.</summary>
        /// <param name="strSoldierName">The player's name.</param>
        public void OnPlayerLeft(String strSoldierName)
        {
            if (mstRunning)
                addToActionQueue(ON_PLAYER_LEAV, strSoldierName);
        }

        /// <summary>Is called when a player changes squads or teams.</summary>
        /// <param name="strSoldierName">The player who changed squads.</param>
        /// <param name="iTeamID">The team the player changed to.</param>
        /// <param name="iSquadID">The squad the player changed to.</param>
        public void OnPlayerSquadChange(String strSoldierName, Int32 iTeamID, Int32 iSquadID)
        {
            if (mstRunning)
                addToActionQueue(ON_PLAYER_TMSQ, strSoldierName, iTeamID, iSquadID);
        }

        /// <summary>Is called when a player list is received.</summary>
        /// <param name="lstPlayers">The list of players.</param>
        /// <param name="cpsSubset">The subset the list of players is of.</param>
        public void OnListPlayers(List<CPlayerInfo> lstPlayers, CPlayerSubset cpsSubset)
        {
            if (mstRunning)
                addToActionQueue(OBTAIN_BC_INFO, lstPlayers);
        }

        /// <summary>Is called when a single player's punkbuster info is received.</summary>
        /// <param name="cpbiPlayer">The player's punkbuster info.</param>
        public void OnPunkbusterPlayerInfo(CPunkbusterInfo cpbiPlayer)
        {
            if (mstRunning)
                addToActionQueue(MANAGE_PB_INFO, cpbiPlayer);
        }

        #endregion

        /// <summary>Constructor.</summary>
        public TeamspeakSync()
        {
            ts3Connection = new TeamspeakConnection(uiTs3ServerIp, uiTs3QueryPort);
            ts3Port = uiTs3ServerPort;
            ts3Username = uiTs3QueryUsername;
            ts3Password = uiTs3QueryPassword;
            ts3StagingChannel = new TeamspeakChannel();
            ts3StagingChannel.tsName = uiTs3StgChannelName;
            ts3TeamChannels = new Dictionary<Int32, TeamspeakChannel>();
            ts3SquadChannels = new Dictionary<Int32, Dictionary<Int32, TeamspeakChannel>>();
            ts3ClientInfo = new List<TeamspeakClient>();
            ts3Response = new TeamspeakResponse("error id=0 msg=ok");

            synMatchingThreshold = uiMatchingThreshold;
            synUpdateInterval = uiUpdateInterval;
            synTeamBasedThreshold = uiTeamBasedThreshold;
            synSquadBasedThreshold = uiSquadBasedThreshold;
            synTeamChannelNames = uiTeamChannelNames;
            synSquadChannelNames = uiSquadChannelNames;
            synRemoveChannelsOnEmpty = uiRemoveChannelsOnEmpty == enumBoolYesNo.Yes;
            synRemoveClients = uiRemoveClients == enumBoolYesNo.Yes;
            synRemoveClientsWhitelist = new List<Int32>();
            synUpdateThread = new Thread(performSynchronize);

            yllYelling = uiYelling == enumBoolYesNo.Yes;
            yllInterval = uiYellingInterval;
            yllOnJoin = uiYellOnJoin == enumBoolYesNo.Yes;
            yllOnJoinDelay = uiYellOnJoinDelay;
            yllMessage = uiYellMessage;
            yllDuration = uiYellDuration;
            yllOnJoinTimers = new Queue<Timer>();
            yllYellingThread = new Thread(performYelling);

            roeReconnectOnError = uiReconnectOnError == enumBoolYesNo.Yes;
            roeReconnectOnErrorAttempts = uiReconnectOnErrorAttempts;
            roeReconnectOnErrorInterval = uiReconnectOnErrorInterval;

            bc2ClientInfo = new List<BadCompanyClient>();
            bc2PunkbusterInfo = new Dictionary<String, CPunkbusterInfo>();

            mstClientInfo = new List<MasterClient>();
            mstActions = new Queue<ActionEvent>();
            mstRunning = false;
            mstDebug = uiDebug == enumBoolYesNo.Yes;
            mstActionAdded = new AutoResetEvent(false);
            mstActionThread = new Thread(performActions);
            mstActionSemaphore = new Semaphore(0, Int32.MaxValue);
            mstActionMutex = new Mutex();
            mstPlayerLeft = false;
            mstPlayerName = String.Empty;
        }

        #region Control Actions

        /// <summary>Performs actions as long as the plugin is enabled.</summary>
        public void performActions()
        {
            ActionEvent curAction;
            try
            {
                #region Setup Operations

                // Open the connection to teamspeak and check for errors.
                // stealth
                ts3Response = performOpenConnection();
                if (ts3Response.Id != 0) { performError(ts3Response); return; }

                // Force Punkbuster information to be sent to PRoCon, then wait a second for it all to arrive.
                // After waiting for a second, force the player list to be sent to PRoCon.
                forcePbListing();
                Thread.Sleep(1000);
                forceBcListing();

                // Start the thread that will notify us at set intervals when to update.
                synUpdateThread = new Thread(performSynchronize);
                synUpdateThread.Start();
                // Start the thread that will notify us at set intervals when to yell.
                yllYellingThread = new Thread(performYelling);
                yllYellingThread.Start();

                #endregion
                #region Perform Actions, Rinse, Repeat

                consoleWrite("Beginning Synchronization.");
                // Only run the plugin while we're supposed to be running it.
                // Only perform an action if there is an action to perform.
                while (true)
                {
                    // Ask the thread to wait until we receive an action to process.
                    mstActionSemaphore.WaitOne();
                    if (!mstRunning)
                        break;

                    // Do not allow the queue to be altered while we're dequeueing an action.
                    mstActionMutex.WaitOne();

                    // Shouldn't even happen, but sometimes when the plugin is warming up, it will
                    // try to dequeue and action when it doesn't have one.
                    if (mstActions.Count <= 0)
                        continue;

                    curAction = mstActions.Dequeue();
                    mstActionMutex.ReleaseMutex();

                    sendQueryAttempt();

                    // Perform the command.
                    switch (curAction.Command)
                    {
                        #region Obtain/Update Information

                        case OBTAIN_TS_INFO:
                            obtainTsInfo();
                            if (ts3Response.Id != 0) { performError(ts3Response); }
                            break;

                        case OBTAIN_BC_INFO:
                            obtainBcInfo((List<CPlayerInfo>)curAction.Argument);
                            break;

                        case MANAGE_PB_INFO:
                            managePbInfo((CPunkbusterInfo)curAction.Argument);
                            break;

                        case UPDATE_MC_INFO:
                            updateMasterInfo();
                            break;

                        #endregion
                        #region Swap/Remove/Yell Actions

                        case CHECK_FOR_SWAP:
                            checkClientForSwap((MasterClient)curAction.Argument);
                            if (ts3Response.Id != 0) { performError(ts3Response); }
                            break;

                        case CHECK_FOR_REMV:
                            checkClientForRemove((MasterClient)curAction.Argument);
                            if (ts3Response.Id != 0) { performError(ts3Response); }
                            break;

                        case CHECK_FOR_YELL:
                            checkClientForYell((MasterClient)curAction.Argument);
                            break;

                        case CHECK_SWAP_ALL:
                            // DEBUG STUFF -------------------------------------------------
                            debugWrite("----------Client Listing (Check Swap All):");
                            foreach (MasterClient mstClient in mstClientInfo)
                                debugWrite(String.Format("----------TS Client [Ip: {0}, Channel: {1}, Name: {2}] / BC Client [Ip: {3}, Team: {4}, Name: {5}]",
                                            (mstClient.HasTsClient) ? mstClient.TsClient.advIpAddress : "Null_IP",
                                            (mstClient.HasTsClient) ? mstClient.TsClient.medChannelId.Value.ToString() : "Null_Channel",
                                            (mstClient.HasTsClient) ? mstClient.TsClient.tsName : "Null_Name",
                                            (mstClient.HasBcClient) ? ((mstClient.BcClient.HasPbInfo) ? mstClient.BcClient.IP : "Null_IP") : "Null_IP",
                                            (mstClient.HasBcClient) ? mstClient.BcClient.TeamId.ToString() : "Null_Team",
                                            (mstClient.HasBcClient) ? mstClient.BcClient.Name : "Null_Name"));
                            // END DEBUG STUFF ---------------------------------------------
                            foreach (MasterClient mstClient in getPlayersOnBothServers())
                                addToActionQueue(CHECK_FOR_SWAP, mstClient);
                            break;

                        case CHECK_REMV_ALL:
                            foreach (MasterClient mstClient in getPlayersOnTsOnly())
                                addToActionQueue(CHECK_FOR_REMV, mstClient);
                            break;

                        case CHECK_YELL_ALL:
                            foreach (MasterClient mstClient in getPlayersOnBcOnly())
                                addToActionQueue(CHECK_FOR_YELL, mstClient);
                            break;

                        case CHECK_YELL_JIN:
                            String Name = (String)curAction.Argument;
                            foreach (MasterClient mstClient in getPlayersOnBcOnly())
                                if (mstClient.BcClient.Name == Name)
                                {
                                    checkClientForYell(mstClient);
                                    break;
                                }
                            break;

                        #endregion
                        #region In-Game Events.

                        case ON_PLAYER_JOIN:
                            playerJoined((String)curAction.Argument);
                            break;

                        case ON_PLAYER_LEAV:
                            mstPlayerName = (String)curAction.Argument;
                            playerLeft(mstPlayerName);
                            break;

                        case ON_PLAYER_TMSQ:
                            playerSwappedTeamsOrSquads((String)curAction.Argument, (Int32)curAction.Argument, (Int32)curAction.Argument);
                            break;

                        #endregion
                    }

                    // For reasoning, see 'playerSwappedTeamsOrSquads'.
                    // Record whether the current event was an ON_PLAYER_LEAV event.
                    mstPlayerLeft = curAction.Command == ON_PLAYER_LEAV;
                }
                consoleWrite("Synchronization Haulted.");

                #endregion
                #region Cleanup Operations

                // Close the connection.
                ts3Connection.closeConnection();

                #endregion
            }
            catch (Exception e)
            {
                #region Exception Handling

                consoleWrite("^8Fatal Error!");
                consoleWrite("^8^bMessage:^n^0 " + e.Message);
                consoleWrite("^8^bStack Trace:^n^0 " + e.StackTrace);
                setPluginState(false);
                consoleWrite("Synchronization Haulted.");

                // Close the connection.
                ts3Connection.closeConnection();

                #endregion
            }
        }

        /// <summary>Handles what happens when a teamspeak error occurs.</summary>
        // stealth ts3ErrorResponse
        public void performError(TeamspeakResponse ts3ErrorResponse)
        {
            // If id < 0, it is a connection issue and the connection has been closed.
            // if id > 0, it is a query error and needs to be handled appropriately.
            // 3329 = You got temp banned from the teamspeak 3 query server because your ip is not whitelisted.
            // 3331 = You are still temp banned from the server.

            // stealth
            if (ts3ErrorResponse.Id < 0)
            {
                consoleWrite("^8An error occurred with the Teamspeak3 connection:");
                consoleWrite(String.Format("^8{0}: {1}", ts3ErrorResponse.Id, ts3ErrorResponse.Message));
                if (roeReconnectOnError)
                    if (performReconnect())
                        return; // Yay! Reconnect succeeded.
                consoleWrite("^8Disabling the plugin.");
                setPluginState(false);
            }
            else if (ts3ErrorResponse.Id == 3331 || ts3ErrorResponse.Id == 3329)
            {
                consoleWrite("^8Make sure your IP is in your Teamspeak3 server's whitelist:");
                consoleWrite(String.Format("^8{0}: {1}", ts3ErrorResponse.Id, ts3ErrorResponse.Message));
                consoleWrite("^8Disabling the plugin.");
                setPluginState(false);
            }
            else
            {
                consoleWrite("^3A query error occurred:");
                consoleWrite(String.Format("^3{0}: {1}", ts3ErrorResponse.Id, ts3ErrorResponse.Message));
            }
        }

        /// <summary>Opens a connection to the Teamspeak Server and obtains all necessary server information.</summary>
        // stealth - void > TeamspeakResponse
        public TeamspeakResponse performOpenConnection()
        {
            // stealth - local var
            TeamspeakResponse ts3ConnectionResponse = new TeamspeakResponse("error id=0 msg=ok");

            consoleWrite("Attempting to establish a connection to Teamspeak.  Please allow 30-60 seconds for this to complete!");
            #region Wait ( 10 seconds ) for data to be initialized.

            // This is only useful if the plugin is enabled when PRoCon starts.
            Int32 secondsSlept = 0;
            while (secondsSlept < 10 && ts3Connection.IP == "Teamspeak Ip")
            {
                secondsSlept++;
                Thread.Sleep(1000);
            }

            #endregion

            #region Attempt to establish a connection

            // stealth
            ts3ConnectionResponse = ts3Connection.openConnection();
            if (ts3ConnectionResponse.Id != 0)
            {
                consoleWrite(String.Format("^8There was an issue establishing a connection to {0}:{1}.", ts3Connection.IP, ts3Connection.Port));
                consoleWrite("^8Check your ^bTeamspeak IP^n and ^bTeamspeak Query Port^n then try again.");
                return new TeamspeakResponse("error id=0 msg=ok");
            }
            consoleWrite(String.Format("^2Established connection to {0}:{1}.", ts3Connection.IP, ts3Connection.Port));

            #endregion
            #region Attempt to login

            // stealth
            ts3ConnectionResponse = ts3Connection.sendQuery(TeamspeakQuery.buildLoginQuery(ts3Username, ts3Password));
            if (ts3ConnectionResponse.Id != 0)
            {
                consoleWrite(String.Format("^8There was an issue logging in as {0}.", ts3Username));
                consoleWrite("^8Check your ^bUsername^n and ^bPassword^n then try again.");
                return new TeamspeakResponse("error id=0 msg=ok"); ;
            }
            consoleWrite(String.Format("^2Authenticated as {0}.", ts3Username));

            #endregion
            #region Attempt to use correct server

            // stealth
            ts3ConnectionResponse = ts3Connection.sendQuery(TeamspeakQuery.buildUsePortQuery(ts3Port));
            if (ts3ConnectionResponse.Id != 0)
            {
                consoleWrite(String.Format("^8There was an issue selecting the virtual server on port {0}.", ts3Port));
                consoleWrite("^8Check your ^bTeamspeak Port^n then try again.");
                return new TeamspeakResponse("error id=0 msg=ok"); ;
            }
            consoleWrite(String.Format("^2Selected the virtual server on port {0}.", ts3Port));

            #endregion
            #region Attempt to find the staging channel

            // stealth
            ts3ConnectionResponse = ts3Connection.sendQuery(TeamspeakQuery.buildChannelFindQuery(ts3StagingChannel.tsName));
            if (ts3ConnectionResponse.Id != 0)
            {
                consoleWrite(String.Format("^8There was an issue finding the channel {0}.", ts3StagingChannel.tsName));
                consoleWrite("^8Check your ^bStaging Channel Name^n then try again.");
                return new TeamspeakResponse("error id=0 msg=ok"); ;
            }
            ts3StagingChannel.setBasicData(ts3ConnectionResponse.Sections[0].Groups[0]);

            consoleWrite("^2Found the staging channel.");

            #endregion

            // Hardcoded to change name to Odin
            TeamspeakQuery changeNick = new TeamspeakQuery("clientupdate");
            changeNick.addParameter("client_nickname", "Odin");
            ts3Connection.sendQuery(changeNick);

            // stealth - wait 3 seconds / return response
            Thread.Sleep(3000);
            return ts3ConnectionResponse;
        }

        /// <summary>Requests a teamspeak and master client update as well as a check for all swap and all remove.</summary>
        public void performSynchronize()
        {
            while (mstRunning)
            {
                addToActionQueue(OBTAIN_TS_INFO);
                addToActionQueue(UPDATE_MC_INFO);
                addToActionQueue(CHECK_SWAP_ALL);
                if (synRemoveClients)
                    addToActionQueue(CHECK_REMV_ALL);
                Thread.Sleep(synUpdateInterval);
            }
        }

        /// <summary>Yells to clients not in teamspeak in intervals.</summary>
        public void performYelling()
        {
            while (mstRunning)
            {
                if (yllYelling)
                    addToActionQueue(CHECK_YELL_ALL);
                Thread.Sleep(yllInterval);
            }
        }

        /// <summary>Attempts to reconnect to the teamspeak server.</summary>
        /// <returns>Whether the reconnect was successful or not.</returns>
        public Boolean performReconnect()
        {
            Int32 attempts = 0;
            while (mstRunning && (roeReconnectOnErrorAttempts == 0 || attempts < roeReconnectOnErrorAttempts))
            {
                attempts++;
                // stealth
                ts3Response = performOpenConnection();
                if (ts3Response.Id == 0)
                    return true;
                consoleWrite(String.Format("^8Failed attempt {0} out of {1}.", attempts, roeReconnectOnErrorAttempts));
                Thread.Sleep(roeReconnectOnErrorInterval);
            }
            return false;
        }

        /// <summary>Destroys any data the plugin may have collected.  I.e. Client info, Actions, etc.</summary>
        public void performCleaning()
        {
            ts3TeamChannels.Clear();
            ts3SquadChannels.Clear();
            ts3ClientInfo.Clear();
            bc2ClientInfo.Clear();
            bc2PunkbusterInfo.Clear();
            mstClientInfo.Clear();
            mstActions.Clear();
        }

        #endregion
        #region Update Information

        /// <summary>Obtain the current Teamspeak Client list for the Bad Company channels.</summary>
        public void obtainTsInfo()
        {
            // stealth
            #region Obtain Basic Information

            queryQueue.Enqueue(new QQItem(TeamspeakQuery.buildClientListQuery(), "clientlist"));

            #endregion

            #region Build Channel List

            queryQueue.Enqueue(new QQItem(TeamspeakQuery.buildChannelListQuery(), "channellist"));

            #endregion
        }

        /// <summary>Obtain the current Bad Company 2 Client list for the players on Bad Company 2.</summary>
        /// <param name="bc2GeneralInfo">The current list of players on bad company 2.</param>
        public void obtainBcInfo(List<CPlayerInfo> bc2GeneralInfo)
        {
            // [0] Used to hold the current client info.
            List<BadCompanyClient> bc2ClientInfo = new List<BadCompanyClient>();

            #region Build Client List with General Information

            // Create Bad Company Clients out of the general information.
            foreach (CPlayerInfo genInfo in bc2GeneralInfo)
                bc2ClientInfo.Add(new BadCompanyClient(genInfo));

            #endregion
            #region Match Clients with Punkbuster Information

            // Only re-store the used punkbuster information.
            Dictionary<String, CPunkbusterInfo> bc2NewPunkbusterInfo = new Dictionary<String, CPunkbusterInfo>();

            // Check if we have punkbuster information for each Bad Company Client.
            foreach (BadCompanyClient bcClient in bc2ClientInfo)
                if (bc2PunkbusterInfo.ContainsKey(bcClient.Name))
                {
                    bcClient.PunkbusterInfo = bc2PunkbusterInfo[bcClient.Name];
                    bc2NewPunkbusterInfo.Add(bcClient.Name, bc2PunkbusterInfo[bcClient.Name]);
                }

            // Remove unused punkbuster information
            bc2PunkbusterInfo = bc2NewPunkbusterInfo;

            #endregion

            // Replace global list with new list.
            this.bc2ClientInfo = bc2ClientInfo;
        }

        /// <summary>Updates the punkbuster information with new information.</summary>
        /// <param name="punkInfo">The punkbuster information to add.</param>
        public void managePbInfo(CPunkbusterInfo punkInfo)
        {
            // Remove the port off the IP of the punkbuster info.
            punkInfo = new CPunkbusterInfo(
                punkInfo.SlotID, punkInfo.SoldierName, punkInfo.GUID,
                punkInfo.Ip.Remove(punkInfo.Ip.IndexOf(':')),
                punkInfo.PlayerCountry, punkInfo.PlayerCountryCode);

            // If it's in the list already, just replace it.  Otherwise, add it to the list.
            if (bc2PunkbusterInfo.ContainsKey(punkInfo.SoldierName))
                bc2PunkbusterInfo[punkInfo.SoldierName] = punkInfo;
            else
                bc2PunkbusterInfo.Add(punkInfo.SoldierName, punkInfo);
        }

        /// <summary>Builds a new master list using the information obtained by other methods.</summary>
        public void updateMasterInfo()
        {
            // [0] Used to hold the current client info.
            // [1] Used to hold the Bad Company Client Master Clients who could not be matched.
            // [2] Used to hold the clients whose IP matched the Bad Company Client's IP
            List<MasterClient> mstClientInfo = new List<MasterClient>();
            List<MasterClient> mstNewClientInfo = new List<MasterClient>();
            List<MasterClient> matchOnIp = new List<MasterClient>();

            #region Build Master List with Teamspeak Clients

            // Create Master Clients out of teamspeak clients.
            foreach (TeamspeakClient tsClient in ts3ClientInfo)
                mstClientInfo.Add(new MasterClient(tsClient));

            #endregion

            // First, determine if the bad company client has an IP.
            foreach (BadCompanyClient bcClient in bc2ClientInfo)
                #region Match Bad Company Clients (with IPs) to Master Clients

                if (bcClient.HasPbInfo)
                {
                    // Second, build a list of clients from the master list whom IP matches this bc client.
                    matchOnIp.Clear();
                    foreach (MasterClient mstClient in mstClientInfo)
                        if (bcClient.IP == mstClient.TsClient.advIpAddress)
                            matchOnIp.Add(mstClient);

                    #region Multiple IPs Match

                    // If the list contains more than one match, match to the client whose name is the closest to this bc client.
                    if (matchOnIp.Count > 1)
                    {
                        Int32 matchIndex = 0;
                        Double biggestPercent = 0;
                        Double currentPercent = 0;
                        for (int i = 0; i < matchOnIp.Count; i++)
                        {
                            currentPercent = MasterClient.calcPercentMatch(bcClient.Name, matchOnIp[i].TsClient.tsName);
                            if (currentPercent > biggestPercent)
                            {
                                matchIndex = i;
                                biggestPercent = currentPercent;
                            }
                        }

                        // The client at index 'matchIndex' had the highest match percentage.
                        // Since this had a previous IP match, we don't care if the name match meets the threshold.
                        matchOnIp[matchIndex].BcClient = bcClient;
                    }

                    #endregion
                    #region Single IP Match

                    // If the list only contains one match, then there were no duplicate IPs, so this is the client.
                    else if (matchOnIp.Count == 1)
                        matchOnIp[0].BcClient = bcClient;

                    #endregion
                    #region No IPs Match

                    // If the list contains no matches, then they are not in teamspeak.
                    else
                        mstNewClientInfo.Add(new MasterClient(bcClient));

                    #endregion

                }

                #endregion
                #region Match Bad Company Clients (without IPs) to Master Clients

                else
                {

                    // There is no IP to match with, so match based on name.
                    Int32 matchIndex = 0;
                    Double biggestPercent = 0;
                    Double currentPercent = 0;
                    for (int i = 0; i < mstClientInfo.Count; i++)
                        if (!mstClientInfo[i].HasBcClient)
                        {
                            currentPercent = MasterClient.calcPercentMatch(bcClient.Name, mstClientInfo[i].TsClient.tsName);
                            if (currentPercent > biggestPercent)
                            {
                                matchIndex = i;
                                biggestPercent = currentPercent;
                            }
                        }

                    // The client at index 'matchIndex' had the highest match percentage.
                    // If our match is above the threshold, the clients match.
                    if (biggestPercent >= synMatchingThreshold)
                        mstClientInfo[matchIndex].BcClient = bcClient;
                    // Otherwise, consider the client not in teamspeak, and add them as BC2 only.
                    else
                        mstNewClientInfo.Add(new MasterClient(bcClient));

                }

                #endregion

            // Consolidate master client lists.
            mstClientInfo.AddRange(mstNewClientInfo);

            // Replace global list with new list.
            this.mstClientInfo = mstClientInfo;
        }

        #endregion
        #region Check Information

        /// <summary>Moves a client into the correct channel, if they're in the wrong channel.</summary>
        /// <param name="client">The client to check.</param>
        public void checkClientForSwap(MasterClient client)
        {
            // Replace current response with "ok" response, so the program doesn't see
            // an issue if we have to bail out because the client isn't in one of the servers.
            ts3Response = new TeamspeakResponse("error id=0 msg=ok");

            // Do not proceed if the client is not in either server.
            if (!client.HasBcClient || !client.HasTsClient)
                return;

            // Save the channel's Id so we can remove the channel if needed.
            Int32? oldChannelId = client.TsClient.medChannelId;

            // Move player to staging channel if team-based swapping is off or there are not enough players (for team-based swapping).
            // If by all means we should be doing team-based swapping, but they are not in a team (spectator?), put them in the staging channel.
            if (synTeamBasedThreshold == -1 || getPlayersOnBothServers().Count < synTeamBasedThreshold || client.BcClient.TeamId == 0)
            {
                #region Move To Staging Channel

                if (client.TsClient.medChannelId != ts3StagingChannel.tsId)
                {
                    debugWrite(String.Format("swapClient - Staging Mode - Swapping Client [{0}] to Channel [{1}]", client.TsClient.tsName, ts3StagingChannel.tsName));
                    // DEBUG STUFF -------------------------------------------------
                    debugWrite("-----------Team Channel Listing:");
                    foreach (KeyValuePair<Int32, TeamspeakChannel> tsPair in ts3TeamChannels)
                        debugWrite(String.Format("-----------For [Team {0}] -> To Channel [Name: {1}, Id: {2}]", tsPair.Key, tsPair.Value.tsName, tsPair.Value.tsId.Value));
                    // END DEBUG STUFF ---------------------------------------------
                    // stealth
                    List<Object> tmp = new List<Object>(); tmp.Add(client);
                    queryQueue.Enqueue(new QQItem(TeamspeakQuery.buildClientMoveQuery(client.TsClient.tsId.Value, ts3StagingChannel.tsId.Value), "swapclient", tmp));
                }

                #endregion
            }
            // Move player to team-based channels if squad-based swapping is off or there are not enough players (for squad-based swapping).
            // If by all means we should be doing squad-based swapping, but they are not in a squad (squad 0), put them in the team channel.
            else if (synSquadBasedThreshold == -1 || getPlayersOnBothServers().Count < synSquadBasedThreshold || client.BcClient.SquadId == 0)
            {
                #region Create/Find Channel

                // Make sure the Team channel exists.
                if (!ts3TeamChannels.ContainsKey(client.BcClient.TeamId))
                    if (!createTeamChannel(client.BcClient.TeamId))
                        return;

                #endregion
                #region Move To Player To Team

                if (client.TsClient.medChannelId != ts3TeamChannels[client.BcClient.TeamId].tsId)
                {
                    debugWrite(String.Format("swapClient - Team Mode (or Squad 0) - Swapping Client [{0}] to Channel [{1}]", client.TsClient.tsName, ts3TeamChannels[client.BcClient.TeamId].tsName));
                    // DEBUG STUFF -------------------------------------------------
                    debugWrite("-----------Team Channel Listing:");
                    foreach (KeyValuePair<Int32, TeamspeakChannel> tsPair in ts3TeamChannels)
                        debugWrite(String.Format("-----------For [Team {0}] -> To Channel [Name: {1}, Id: {2}]", tsPair.Key, tsPair.Value.tsName, tsPair.Value.tsId.Value));
                    // END DEBUG STUFF ---------------------------------------------
                    // stealth
                    List<Object> tmp = new List<Object>(); tmp.Add(client);
                    queryQueue.Enqueue(new QQItem(TeamspeakQuery.buildClientMoveQuery(client.TsClient.tsId.Value, ts3TeamChannels[client.BcClient.TeamId].tsId.Value), "teamswapclient", tmp));
                }

                #endregion
            }
            // Move player to squad-based channels.
            else
            {
                #region Create/Find Channel

                // Make sure the Team channel exists.
                if (!ts3TeamChannels.ContainsKey(client.BcClient.TeamId))
                    if (!createTeamChannel(client.BcClient.TeamId))
                        return;
                // Make sure the Squad channel exists.
                if (!ts3SquadChannels[client.BcClient.TeamId].ContainsKey(client.BcClient.SquadId))
                    if (!createSquadChannel(client.BcClient.TeamId, client.BcClient.SquadId))
                        return;

                #endregion
                #region Move To Player To Squad

                if (client.TsClient.medChannelId != ts3SquadChannels[client.BcClient.TeamId][client.BcClient.SquadId].tsId)
                {
                    debugWrite(String.Format("swapClient - Squad Mode - Swapping Client [{0}] to Channel [{1}]", client.TsClient.tsName, ts3SquadChannels[client.BcClient.TeamId][client.BcClient.SquadId].tsName));
                    // DEBUG STUFF -------------------------------------------------
                    debugWrite("-----------Team Channel Listing:");
                    foreach (KeyValuePair<Int32, TeamspeakChannel> tsPair in ts3TeamChannels)
                        debugWrite(String.Format("-----------For [Team {0}] -> To Channel [Name: {1}, Id: {2}]", tsPair.Key, tsPair.Value.tsName, tsPair.Value.tsId.Value));
                    // END DEBUG STUFF ---------------------------------------------
                    // stealth
                    List<Object> tmp = new List<Object>(); tmp.Add(client);
                    queryQueue.Enqueue(new QQItem(TeamspeakQuery.buildClientMoveQuery(client.TsClient.tsId.Value, ts3SquadChannels[client.BcClient.TeamId][client.BcClient.SquadId].tsId.Value), "squadswapclient", tmp));
                    
                }

                #endregion
            }

            // Delete the channel if the remove option is enabled.
            if (synRemoveChannelsOnEmpty)
                deleteChannel(oldChannelId);

        }

        /// <summary>Removes a client from the attacking/defending channel, if they need to be.</summary>
        /// <param name="client">The client to check.</param>
        public void checkClientForRemove(MasterClient client)
        {
            // Replace current response with "ok" response, so the program doesn't see
            // an issue if we have to bail out because the client isn't in one of the servers.
            ts3Response = new TeamspeakResponse("error id=0 msg=ok");

            // Do not proceed if the client is in bad company 2 or not in teamspeak.
            if (client.HasBcClient || !client.HasTsClient)
                return;

            // Save the channel's Id so we can remove the channel if needed.
            Int32? oldChannelId = client.TsClient.medChannelId;

            // Only attempt to remove the client if they are not in the whitelist.
            if (!synRemoveClientsWhitelist.Contains(client.TsClient.medDatabaseId.Value))
            {
                Boolean isInTeamOrSquadChannel = false;

                #region Determine If Player Is In Team Or Squad Channel

                // Check to see if the client is in a team channel.
                foreach (Int32 teamId in ts3TeamChannels.Keys)
                    if (ts3TeamChannels[teamId].tsId == client.TsClient.medChannelId)
                    {
                        isInTeamOrSquadChannel = true;
                        break;
                    }
                // Check to see if the client is in a squad channel.
                foreach (Int32 teamId in ts3TeamChannels.Keys)
                {
                    foreach (Int32 squadId in ts3SquadChannels[teamId].Keys)
                        if (ts3SquadChannels[teamId][squadId].tsId == client.TsClient.medChannelId)
                        {
                            isInTeamOrSquadChannel = true;
                            break;
                        }
                    if (isInTeamOrSquadChannel)
                        break;
                }

                #endregion
                #region Move To Staging Area

                // Remove them from the team/squad channel.
                if (isInTeamOrSquadChannel)
                {
                    debugWrite(String.Format("removeClient - Removing Client [{0}]", client.TsClient.tsName));
                    // stealth
                    List<Object> tmp = new List<Object>(); tmp.Add(client);
                    queryQueue.Enqueue(new QQItem(TeamspeakQuery.buildClientMoveQuery(client.TsClient.tsId.Value, ts3StagingChannel.tsId.Value), "removeclient", tmp));
                }

                #endregion
            }


            // Delete the channel if the remove option is enabled.
            if (synRemoveChannelsOnEmpty)
                deleteChannel(oldChannelId);
        }

        /// <summary>Yells to a client if they aren't in teamspeak.</summary>
        /// <param name="client">The client to check.</param>
        public void checkClientForYell(MasterClient client)
        {
            // Do not proceed if the client is in teamspeak or not in bad company 2.
            if (client.HasTsClient || !client.HasBcClient)
                return;

            debugWrite(String.Format("yellClient - Yelling to Client [{0}]", client.BcClient.Name));
            yellToPlayer(yllMessage, yllDuration, client.BcClient.Name);
        }

        #endregion
        #region Bad Company 2 Events

        /// <summary>Yells at a client after a set delay if they aren't on teamspeak.</summary>
        /// <param name="Name">The name of the client who just joined.</param>
        public void playerJoined(String Name)
        {
            debugWrite(String.Format("eventJoin - [{0}] joined the Bad Company 2 server", Name));

            if (yllYelling && yllOnJoin)
                yllOnJoinTimers.Enqueue(new Timer(new TimerCallback(yellOnJoinCallback), Name, yllOnJoinDelay, Timeout.Infinite));
        }

        /// <summary>Removes the client from the bad company client list as well as the master client list.</summary>
        /// <param name="Name">The name of the player that left bad company 2.</param>
        public void playerLeft(String Name)
        {
            debugWrite(String.Format("eventLeave - [{0}] left the Bad Company 2 server", Name));

            #region Remove from Bad Company client list

            foreach (BadCompanyClient bcClient in bc2ClientInfo)
                if (bcClient.Name == Name)
                {
                    bc2ClientInfo.Remove(bcClient);
                    break;
                }

            #endregion
            #region Remove from Master client list

            foreach (MasterClient mstClient in mstClientInfo)
                if (mstClient.HasBcClient && mstClient.BcClient.Name == Name)
                {
                    mstClient.BcClient = null;
                    if (!mstClient.HasTsClient)
                        mstClientInfo.Remove(mstClient);
                    else
                        addToActionQueue(CHECK_FOR_REMV, mstClient);
                    break;
                }

            #endregion
        }

        /// <summary>Updates the client's Team/Squad information and swaps them if neccessary.</summary>
        /// <param name="Name">The name of the client that swapped.</param>
        /// <param name="TeamId">The team the client swapped to.</param>
        /// <param name="SquadId">The squad the client swapped to.</param>
        public void playerSwappedTeamsOrSquads(String Name, Int32 TeamId, Int32 SquadId)
        {
            // The swap event is called after a player leaves the server for some reason.
            // So to counteract this, check to see if the previous action was this player
            // leaving the server.  If it was this player, exit the method.
            if (mstPlayerLeft && mstPlayerName == Name)
                return;

            // Actually handle real player swapped events.
            debugWrite(String.Format("eventSwap - [{0}] swapped to [Team {1}, Squad {2}]", Name, TeamId, SquadId));
            Boolean wasInServer = false;

            #region If Player Was Already In Server

            // Note: Will not find the player if they just joined.
            // The player will be found when new bc2 client information is
            // sent to PRoCon.
            foreach (MasterClient mstClient in mstClientInfo)
                if (mstClient.HasBcClient && mstClient.BcClient.Name == Name)
                {
                    // Update the master client list and the bc2 client list with new bad company client information.
                    mstClient.BcClient = new BadCompanyClient(new CPlayerInfo(mstClient.BcClient.Name, mstClient.BcClient.Tags, TeamId, SquadId));
                    foreach (BadCompanyClient bc2Client in bc2ClientInfo)
                        if (bc2Client.Name == Name)
                        {
                            bc2ClientInfo.Remove(bc2Client);
                            break;
                        }
                    bc2ClientInfo.Add(mstClient.BcClient);

                    // Check for swap if they're in teamspeak.
                    if (mstClient.HasTsClient)
                        addToActionQueue(CHECK_FOR_SWAP, mstClient);

                    // Done.
                    wasInServer = true;
                    break;
                }

            #endregion
            #region If Player Just Joined Server

            // Note: Attempts to match on name to avoid requesting PB info
            // more than necessary.  Should possibly request PB info in the future.
            if (!wasInServer)
            {
                // On certain occasions, the player may already be in the BC2 client list.
                // If they are in the list, update them (a.k.a. remove them), otherwise, add them to it.
                BadCompanyClient bcClient = new BadCompanyClient(new CPlayerInfo(Name, null, TeamId, SquadId));
                foreach (BadCompanyClient bc2Client in bc2ClientInfo)
                    if (bc2Client.Name == Name)
                    {
                        bc2ClientInfo.Remove(bc2Client);
                        break;
                    }
                bc2ClientInfo.Add(bcClient);

                // Attempt to match the bad company client based on their name to a teamspeak client.
                MasterClient matchClient = null;
                Double biggestPercent = 0;
                Double currentPercent = 0;
                foreach (MasterClient mstClient in getPlayersOnTsOnly())
                {
                    currentPercent = MasterClient.calcPercentMatch(bcClient.Name, mstClient.TsClient.tsName);
                    if (currentPercent > biggestPercent)
                    {
                        matchClient = mstClient;
                        biggestPercent = currentPercent;
                    }
                }

                // If our match is above the threshold, the clients match.
                // Update the Master client list then check the player for a swap.
                if (biggestPercent >= synMatchingThreshold && matchClient != null)
                {
                    matchClient.BcClient = bcClient;
                    addToActionQueue(CHECK_FOR_SWAP, matchClient);
                }
                // Otherwise, consider the client not in teamspeak, and add them as BC2 only.
                else
                    mstClientInfo.Add(new MasterClient(bcClient));
            }

            #endregion
        }

        #endregion
        #region Helper Methods

        #region Channel Manager Helpers

        /// <summary>Attempts to create or find the channel specified.</summary>
        /// <param name="TeamId">The team to create/find a channel for.</param>
        public Boolean createTeamChannel(Int32 TeamId)
        {
            // Object to hold the information about the channel for this team.
            TeamspeakChannel newChannel = new TeamspeakChannel();
            debugWrite(String.Format("createTeam - Create channel for [Team {0}]", TeamId));

            // Channel Name - If the user supplied enough names for this team, then use what
            // the user supplied.  Otherwise, use the default name for this team's channel.
            if (synTeamChannelNames.Length >= TeamId && TeamId > 0)
                newChannel.tsName = synTeamChannelNames[TeamId - 1];
            else
                newChannel.tsName = String.Format("Team {0}", TeamId);
            debugWrite(String.Format("createTeam - Channel Name [{0}]", newChannel.tsName));

            // Query to create the channel.
            TeamspeakQuery createChannel = new TeamspeakQuery("channelcreate");
            createChannel.addParameter("channel_name", newChannel.tsName);
            createChannel.addParameter("channel_password", "rAnD0m_Pa$sWurD");
            createChannel.addParameter("channel_flag_permanent", "1");
            createChannel.addParameter("cpid", ts3StagingChannel.tsId.Value.ToString());

            // Sorting - Determine which channel to place this new channel after.
            // Look through all the team channels and find the channel who's team id is
            // closest to the new channel's team id yet below it.
            if (ts3TeamChannels.Count > 0)
                for (int i = TeamId - 1; i >= -1; i--)
                    if (ts3TeamChannels.ContainsKey(i))
                    {
                        debugWrite(String.Format("createTeam - Order after [Channel Id {0}]", ts3TeamChannels[i].tsId));
                        createChannel.addParameter("channel_order", ts3TeamChannels[i].tsId.ToString());
                        break;
                    }
                    else if (i == -1)
                    {
                        debugWrite(String.Format("createTeam - Order after [Top]"));
                        createChannel.addParameter("channel_order", "0");
                        break;
                    }

            // Attempt to create the channel.
            ts3Response = ts3Connection.sendQuery(createChannel);
            debugWrite(String.Format("createTeam - Sent Channel Create Query"));

            // stealth
            Thread.Sleep(3000);

            #region Handle Response

            // If the channel was already created, attempt to find it. ID 771 = Channel already exists.
            if (ts3Response.Id == 771)
                newChannel.tsId = findSubChannel(newChannel.tsName, ts3StagingChannel.tsId.Value);
            // Manually handle a "response took too long" error (-6).
            else if (ts3Response.Id == -6)
            {
                consoleWrite("^8The response to \"Create Channel\" took too long.  Re-establishing connection.");
                consoleWrite(String.Format("^8{0}: {1}", ts3Response.Id, ts3Response.Message));
                performReconnect();
            }
            // If there was an error in creating the channel, bail out.
            else if (ts3Response.Id != 0 || !ts3Response.HasSections || !ts3Response.Sections[0].HasGroups)
                return false;
            // Otherwise, attempt to parse the information out of the create channel query.
            else
            {
                Int32 intOut;
                if (!Int32.TryParse(ts3Response.Sections[0].Groups[0]["cid"], out intOut))
                    return false;
                newChannel.tsId = intOut;
                debugWrite(String.Format("createTeam - Created Channel [{0}]", newChannel.tsName));
            }

            // Check if the channel Id has been set (incase we had to find the channel).
            if (newChannel.tsId == null)
                return false;

            #endregion

            // Add the new channel to the master channels list.
            ts3TeamChannels.Add(TeamId, newChannel);
            ts3SquadChannels.Add(TeamId, new Dictionary<Int32, TeamspeakChannel>());
            debugWrite(String.Format("createTeam - Added [{0}] to channels", newChannel.tsName));

            return true;
        }

        /// <summary>Attempts to create or find the channel specified.</summary>
        /// <param name="TeamId">The team to create/find a squad channel for.</param>
        /// <param name="SquadId">The squad to create/find a channel for.</param>
        public Boolean createSquadChannel(Int32 TeamId, Int32 SquadId)
        {
            // Object to hold the information about the channel for this team.
            TeamspeakChannel newChannel = new TeamspeakChannel();
            debugWrite(String.Format("createSquad - Create channel for [Team {0}, Squad {1}]", TeamId, SquadId));

            // Channel Name - If the user supplied enough names for this squad, then use what
            // the user supplied.  Otherwise, use the default name for this team's channel.
            if (synSquadChannelNames.Length >= SquadId && SquadId > 0)
                newChannel.tsName = synSquadChannelNames[SquadId - 1];
            else
                newChannel.tsName = String.Format("Squad {0}", SquadId);
            debugWrite(String.Format("createSquad - Channel Name [{0}]", newChannel.tsName));

            // Query to create the channel.
            TeamspeakQuery createChannel = new TeamspeakQuery("channelcreate");
            createChannel.addParameter("channel_name", newChannel.tsName);
            createChannel.addParameter("channel_password", "rAnD0m_Pa$sWurD");
            createChannel.addParameter("channel_flag_permanent", "1");
            createChannel.addParameter("cpid", ts3TeamChannels[TeamId].tsId.Value.ToString());

            // Sorting - Determine which channel to place this new channel after.
            // Look through all this team's squad channels and find the channel who's squad id is
            // closest to the new channel's squad id yet below it.
            if (ts3TeamChannels.Count > 0)
                for (int i = SquadId - 1; i >= -1; i--)
                    if (ts3SquadChannels[TeamId].ContainsKey(i))
                    {
                        debugWrite(String.Format("createSquad - Order after [Channel Id {0}]", ts3SquadChannels[TeamId][i].tsId));
                        createChannel.addParameter("channel_order", ts3SquadChannels[TeamId][i].tsId.ToString());
                        break;
                    }
                    else if (i == -1)
                    {
                        debugWrite(String.Format("createSquad - Order after [Top]"));
                        createChannel.addParameter("channel_order", "0");
                        break;
                    }

            // Attempt to create the channel.
            ts3Response = ts3Connection.sendQuery(createChannel);
            debugWrite(String.Format("createSquad - Sent Channel Create Query"));

            // stealth
            Thread.Sleep(3000);

            #region Handle Response

            // If the channel was already created, attempt to find it. ID 771 = Channel already exists.
            if (ts3Response.Id == 771)
                newChannel.tsId = findSubChannel(newChannel.tsName, ts3TeamChannels[TeamId].tsId.Value);
            // Manually handle a "response took too long" error (-6).
            else if (ts3Response.Id == -6)
            {
                consoleWrite("^8The response to \"Create Channel\" took too long.  Re-establishing connection.");
                consoleWrite(String.Format("^8{0}: {1}", ts3Response.Id, ts3Response.Message));
                performReconnect();
            }
            // If there was an error in creating the channel, bail out.
            else if (ts3Response.Id != 0 || !ts3Response.HasSections || !ts3Response.Sections[0].HasGroups)
                return false;
            // Otherwise, attempt to parse the information out of the create channel query.
            else
            {
                Int32 intOut;
                if (!Int32.TryParse(ts3Response.Sections[0].Groups[0]["cid"], out intOut))
                    return false;
                newChannel.tsId = intOut;
                debugWrite(String.Format("createSquad - Created Channel [{0}]", newChannel.tsName));
            }

            // Check if the channel Id has been set (incase we had to find the channel).
            if (newChannel.tsId == null)
                return false;

            #endregion

            // Add the new channel to the master channels list.
            ts3SquadChannels[TeamId].Add(SquadId, newChannel);
            debugWrite(String.Format("createSquad - Added [{0}] to channels", newChannel.tsName));

            return true;
        }

        /// <summary>Attempts to find a channel with a specified name in a specific channel.</summary>
        /// <param name="channelName">The name of the channel we're trying to find.</param>
        /// <param name="parentId">The Id of the parent channel.</param>
        /// <returns>The Id of the channel we're looking for.  Null if not found.</returns>
        public Int32? findSubChannel(String channelName, Int32 parentId)
        {
            List<TeamspeakChannel> channels = new List<TeamspeakChannel>();
            debugWrite(String.Format("findChannel - Find Channel [{0}] under Parent Id [{1}]", channelName, parentId));

            // Request all the channels be sent back.  Bail out if the query was bad.
            ts3Response = ts3Connection.sendQuery(TeamspeakQuery.buildChannelListQuery());
            // stealth
            Thread.Sleep(3000);
            if (ts3Response.Id != 0)
                return null;
            debugWrite(String.Format("findChannel - Sent Channel List Query"));

            // Build a list of channels.
            foreach (TeamspeakResponseSection section in ts3Response.Sections)
                foreach (TeamspeakResponseGroup group in section.Groups)
                    channels.Add(new TeamspeakChannel(group));

            // Determine which channel is the one we're looking for.
            foreach (TeamspeakChannel tsChannel in channels)
                if (tsChannel.medPId.Value == parentId && tsChannel.tsName == channelName)
                    return tsChannel.tsId;
            debugWrite(String.Format("findChannel - Could NOT find [{0}] under Parent Id [{1}]", channelName, parentId));

            // Didn't find the channel.
            return null;
        }

        /// <summary>Deletes the channel specified (and it's parent if it is a squad channel)
        /// only if the specified channel(s) do not have people in them.  It will not delete
        /// the channel if it is not a team or squad channel.</summary>
        /// <param name="channelId">The channel to attempt to remove.</param>
        public void deleteChannel(Int32? channelId)
        {
            // Attempt to remove the channel.  Make sure we're not removing the staging channel.
            Boolean isSquadChannel = false;
            Int32 oldTeamId = 0;
            Int32 oldSquadId = 0;

            debugWrite(String.Format("deleteChannel - Check Channel Id [{0}] for deletion", channelId));
            #region Determine Which channels To Delete.

            // Check if it is a team channel.
            foreach (Int32 teamId in ts3TeamChannels.Keys)
                if (ts3TeamChannels[teamId].tsId == channelId)
                {
                    debugWrite(String.Format("deleteChannel - Channel Id [{0}] matched to Team Channel [{1}]", channelId, ts3TeamChannels[teamId].tsName));
                    isSquadChannel = false;
                    oldTeamId = teamId;
                    oldSquadId = -1;
                    break;
                }
                else
                    // Check if it is a squad channel.
                    foreach (Int32 squadId in ts3SquadChannels[teamId].Keys)
                        if (ts3SquadChannels[teamId][squadId].tsId == channelId)
                        {
                            debugWrite(String.Format("deleteChannel - Channel Id [{0}] matched to Squad Channel [{1}]", channelId, ts3SquadChannels[teamId][squadId].tsName));
                            isSquadChannel = true;
                            oldTeamId = teamId;
                            oldSquadId = squadId;
                            break;
                        }

            // If the channel could not be found, exit. (Both values can never be 0 if found.)
            if (oldTeamId == 0 && oldSquadId == 0)
                return;

            #endregion

            TeamspeakQuery deleteChannel = new TeamspeakQuery("channeldelete");

            #region Delete Squad Channel

            // If it's a squad channel, attempt to delete all it's siblings.
            // By attempting to delete all it's siblings, this fixes issues where
            //  trying to delete the parent channel would result in all the empty
            //  siblings being deleted in ts3 but not from the list in the code.
            if (isSquadChannel)
            {
                List<Int32> deletedSquadIDs = new List<Int32>();

                // Remove the squads from teamspeak and keep track of the id's to delete from code behind.
                foreach (Int32 squadId in ts3SquadChannels[oldTeamId].Keys)
                {
                    deleteChannel.addParameter("cid", ts3SquadChannels[oldTeamId][squadId].tsId.ToString());
                    ts3Response = ts3Connection.sendQuery(deleteChannel);
                    debugWrite(String.Format("deleteChannel - Sent Channel Delete Query for [Team {0}, Squad {1}]", oldTeamId, squadId));

                    Thread.Sleep(3000);

                    #region Handle Response

                    // Ignore "people are in channel" error (772).
                    if (ts3Response.Id == 772)
                        ts3Response = new TeamspeakResponse("error id=0 msg=ok");
                    // Manually handle a "response took too long" error (-6).
                    else if (ts3Response.Id == -6)
                    {
                        consoleWrite("^8The response to \"Delete Channel\" took too long.  Re-establishing connection.");
                        consoleWrite(String.Format("^8{0}: {1}", ts3Response.Id, ts3Response.Message));
                        performReconnect();
                    }
                    // Bail out if there is a different query error.
                    else if (ts3Response.Id != 0)
                        return;
                    // Otherwise, the removal was successful, so remove the channel from our squad channel list.
                    else
                        deletedSquadIDs.Add(squadId);

                    #endregion

                    deleteChannel.removeParameter("cid");
                }
                // Remove the squads from the code behind.
                foreach (Int32 squadId in deletedSquadIDs)
                {
                    ts3SquadChannels[oldTeamId].Remove(squadId);
                    debugWrite(String.Format("deleteChannel - Successfully Deleted [Squad {0}]", squadId));
                }
            }

            #endregion
            #region Delete Team Channel

            // Attempt to delete the parent channel.
            deleteChannel.addParameter("cid", ts3TeamChannels[oldTeamId].tsId.ToString());
            ts3Response = ts3Connection.sendQuery(deleteChannel);
            debugWrite(String.Format("deleteChannel - Sent Channel Delete Query for [Team {0}]", oldTeamId));

            Thread.Sleep(3000);

            #region Handle Response

            // Ignore "people are in channel" error (772).
            if (ts3Response.Id == 772)
                ts3Response = new TeamspeakResponse("error id=0 msg=ok");
            // Manually handle a "response took too long" error (-6).
            else if (ts3Response.Id == -6)
            {
                consoleWrite("^8The response to \"Delete Channel\" took too long.  Re-establishing connection.");
                consoleWrite(String.Format("^8{0}: {1}", ts3Response.Id, ts3Response.Message));
                performReconnect();
            }
            // Bail out if there is a different query error.
            else if (ts3Response.Id != 0)
                return;
            // Otherwise, the removal was successful, so remove the channel from our channel lists.
            else
            {
                debugWrite(String.Format("deleteChannel - Successfully Deleted [Team {1}]", oldTeamId));
                ts3TeamChannels.Remove(oldTeamId);
                ts3SquadChannels.Remove(oldTeamId);
            }

            #endregion

            #endregion
        }

        #endregion

        /// <summary>Retrieves a list of players that are on both the teamspeak 3 and bad company 2 server.</summary>
        /// <returns>A list of players on both servers.</returns>
        public List<MasterClient> getPlayersOnBothServers()
        {
            List<MasterClient> returnList = new List<MasterClient>();
            foreach (MasterClient masterClient in mstClientInfo)
                if (masterClient.HasBcClient && masterClient.HasTsClient)
                    returnList.Add(masterClient);
            return returnList;
        }

        /// <summary>Retrieves a list of players that are only on the teamspeak 3 server.</summary>
        /// <returns>A list of players only on ts3.</returns>
        public List<MasterClient> getPlayersOnTsOnly()
        {
            List<MasterClient> returnList = new List<MasterClient>();
            foreach (MasterClient masterClient in mstClientInfo)
                if (!masterClient.HasBcClient && masterClient.HasTsClient)
                    returnList.Add(masterClient);
            return returnList;
        }

        /// <summary>Retrieves a list of players that are only on the bad company 2 server.</summary>
        /// <returns>A list of players only on bc2.</returns>
        public List<MasterClient> getPlayersOnBcOnly()
        {
            List<MasterClient> returnList = new List<MasterClient>();
            foreach (MasterClient masterClient in mstClientInfo)
                if (masterClient.HasBcClient && !masterClient.HasTsClient)
                    returnList.Add(masterClient);
            return returnList;
        }

        /// <summary>Retrieves a list of players that are on the teamspeak server.</summary>
        /// <returns>A list of players on the teamspeak server.</returns>
        public List<MasterClient> getPlayersOnTs()
        {
            List<MasterClient> returnList = new List<MasterClient>();
            foreach (MasterClient masterClient in mstClientInfo)
                if (masterClient.HasTsClient)
                    returnList.Add(masterClient);
            return returnList;
        }

        /// <summary>Retrieves a list of players that are on the bad company 2 server.</summary>
        /// <returns>A list of players on the bad company 2 server.</returns>
        public List<MasterClient> getPlayersOnBc()
        {
            List<MasterClient> returnList = new List<MasterClient>();
            foreach (MasterClient masterClient in mstClientInfo)
                if (masterClient.HasBcClient)
                    returnList.Add(masterClient);
            return returnList;
        }

        /// <summary>Adds a command to the action queue with the specified arguments.</summary>
        /// <param name="command">The command to execute.</param>
        /// <param name="arguments">The arguments for this command.</param>
        public void addToActionQueue(Int32 command, params Object[] arguments)
        {
            mstActionMutex.WaitOne();
            mstActions.Enqueue(new ActionEvent(command, arguments));
            mstActionMutex.ReleaseMutex();
            mstActionSemaphore.Release();
        }

        /// <summary>Is called by a timer.  Assumes top of queue is timer that called method.</summary>
        /// <param name="Name">The name of the player that joined.</param>
        public void yellOnJoinCallback(Object Name)
        {
            yllOnJoinTimers.Dequeue().Dispose();
            addToActionQueue(CHECK_YELL_JIN, Name);
        }

        #endregion
    }
}
